import pytest
import warnings

import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDDataArray, StdDevUncertainty
from astropy.table import QTable
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import AstropyUserWarning
from glue.core.roi import CircularROI, RectangularROI
from numpy.testing import assert_allclose, assert_array_equal
from regions import (CirclePixelRegion, CircleAnnulusPixelRegion, EllipsePixelRegion,
                     RectanglePixelRegion, PixCoord)
from specutils import Spectrum1D
from specutils.manipulation import FluxConservingResampler

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               flux_conversion_general)

calspec_url = "https://archive.stsci.edu/hlsps/reference-atlases/cdbs/current_calspec/"


def test_version_after_nddata_update(cubeviz_helper, spectrum1d_cube_with_uncerts):
    # Also test that plugin is disabled before data is loaded.
    plg = cubeviz_helper.plugins['Spectral Extraction']
    assert plg._obj.disabled_msg != ''

    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts)

    spectral_cube = cubeviz_helper.app.data_collection[0].get_object(NDDataArray)
    uncert_cube = cubeviz_helper.app.data_collection[1].get_object(StdDevUncertainty)
    spectral_cube.uncertainty = uncert_cube

    # Collapse the spectral cube using the astropy.nddata machinery.
    # Axes 0, 1 are the spatial ones.
    collapsed_cube_nddata = spectral_cube.sum(axis=(0, 1))  # return NDDataArray

    # when loaded into app, cubes loaded in flux are converted to per-pixel-squared
    # surface brightness, so multiply by pix**2 to compare to NDData, if input
    # cube was in flux
    collapsed_cube_nddata = collapsed_cube_nddata * (u.pix ** 2)

    # Collapse the spectral cube using the methods in jdaviz:
    collapsed_cube_s1d = plg.extract(add_data=False)  # returns Spectrum1D

    assert plg._obj.disabled_msg == ''
    assert isinstance(spectral_cube, NDDataArray)
    assert isinstance(collapsed_cube_s1d, Spectrum1D)

    assert_allclose(
        collapsed_cube_nddata.data,
        collapsed_cube_s1d.flux.to_value(collapsed_cube_nddata.unit)
    )


def test_gauss_smooth_before_spec_extract(cubeviz_helper, spectrum1d_cube_with_uncerts):
    # Also test if gaussian smooth plugin is run before spec extract
    # that spec extract yields results of correct cube data
    gs_plugin = cubeviz_helper.plugins['Gaussian Smooth']._obj

    # give uniform unit uncertainties for spec extract test:
    spectrum1d_cube_with_uncerts.uncertainty = StdDevUncertainty(
        np.ones_like(spectrum1d_cube_with_uncerts.data)
    )

    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts)

    gs_plugin.dataset_selected = f'{cubeviz_helper.app.data_collection[0].label}'
    gs_plugin.mode_selected = 'Spatial'
    gs_plugin.stddev = 3

    with pytest.warns(
            AstropyUserWarning,
            match='The following attributes were set on the data object, but will be ignored'):
        gs_plugin.vue_apply()

    gs_data_label = cubeviz_helper.app.data_collection[2].label
    cubeviz_helper.app.add_data_to_viewer('flux-viewer', gs_data_label)

    # create a subset with a single pixel:
    regions = [
        # create a subset with a single pixel:
        CirclePixelRegion(PixCoord(0, 1), radius=0.7),
        # two-pixel region:
        CirclePixelRegion(PixCoord(0.5, 0), radius=1.2)
    ]
    cubeviz_helper.plugins['Subset Tools'].import_region(regions, combination_mode='new')

    extract_plugin = cubeviz_helper.plugins['Spectral Extraction']
    extract_plugin.function = "Sum"
    expected_uncert = 2

    extract_plugin.aperture = 'Subset 1'
    collapsed_spec = extract_plugin.extract()

    # this single pixel has two wavelengths, and all uncertainties are unity
    # irrespective of which collapse function is applied:
    assert len(collapsed_spec.flux) == 2
    assert_array_equal(collapsed_spec.uncertainty.array, 1)

    # this two-pixel region has four unmasked data points per wavelength:
    extract_plugin.aperture = 'Subset 2'
    collapsed_spec_2 = extract_plugin.extract()
    assert_array_equal(collapsed_spec_2.uncertainty.array, expected_uncert)


@pytest.mark.parametrize(
    ("function, expected_uncert"),
    [("Sum", 2), ("Mean", 0.5), ("Min", 1), ("Max", 1)]
)
def test_subset(
    cubeviz_helper, spectrum1d_cube_with_uncerts, function, expected_uncert
):
    # give uniform unit uncertainties for this test:
    spectrum1d_cube_with_uncerts.uncertainty = StdDevUncertainty(
        np.ones_like(spectrum1d_cube_with_uncerts.data)
    )

    regions = [
        # create a subset with a single pixel:
        CirclePixelRegion(PixCoord(0, 1), radius=0.7),
        # two-pixel region:
        CirclePixelRegion(PixCoord(0.5, 0), radius=1.2)
    ]

    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts)
    cubeviz_helper.plugins['Subset Tools'].import_region(regions, combination_mode='new')

    plg = cubeviz_helper.plugins['Spectral Extraction']
    plg.function = function

    # single pixel region:
    plg.aperture = 'Subset 1'
    collapsed_spec_1 = plg.extract()

    # this single pixel has two wavelengths, and all uncertainties are unity
    # irrespective of which collapse function is applied:
    assert len(collapsed_spec_1.flux) == 2
    assert_array_equal(collapsed_spec_1.uncertainty.array, 1)

    # this two-pixel region has four unmasked data points per wavelength:
    plg.aperture = 'Subset 2'
    collapsed_spec_2 = plg.extract()

    assert_array_equal(collapsed_spec_2.uncertainty.array, expected_uncert)


def test_extracted_file_in_export_plugin(cubeviz_helper, spectrum1d_cube_with_uncerts, tmp_path):

    cubeviz_helper.load_data(spectrum1d_cube_with_uncerts)

    extract_plugin = cubeviz_helper.plugins['Spectral Extraction']

    # make sure export enabled is true, and that before the collapse function
    # is run `collapsed_spec_available` is correctly set to False
    assert extract_plugin._obj.export_enabled
    assert extract_plugin._obj.extraction_available is False

    # run extract function, and make sure `extraction_available` was set to True
    extract_plugin._obj.vue_spectral_extraction()
    assert extract_plugin._obj.extraction_available

    # check that default filename is correct, then change path
    fname = 'extracted_sum_Unknown spectrum object_FLUX.fits'
    fname_path = tmp_path / fname
    assert extract_plugin._obj.filename == fname
    extract_plugin._obj.filename = str(fname_path)

    label = extract_plugin._obj.add_results.label
    export_plugin = cubeviz_helper.plugins['Export']._obj

    assert label in export_plugin.data_collection.labels


def test_aperture_markers(cubeviz_helper, spectrum1d_cube):

    cubeviz_helper.load_data(spectrum1d_cube)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        [CirclePixelRegion(PixCoord(0.5, 0), radius=1.2)])

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']
    slice_plg = cubeviz_helper.plugins['Slice']

    mark = extract_plg.aperture.marks[0]
    assert not mark.visible
    with extract_plg.as_active():
        assert mark.visible
        assert not len(mark.x)

        extract_plg.aperture = 'Subset 1'
        before_x = mark.x
        assert len(before_x) > 0

        # sample cube only has 2 slices with wavelengths [4.62280007e-07 4.62360028e-07] m
        slice_values = [4.62280007e-07, 4.62360028e-07]
        slice_plg.value = slice_values[1]
        assert mark.x[1] == before_x[1]

        slice_plg.value = slice_values[0]
        extract_plg._obj.dev_cone_support = True
        extract_plg._obj.wavelength_dependent = True
        assert mark.x[1] == before_x[1]

        slice_plg.value = slice_values[1]
        assert mark.x[1] != before_x[1]

        extract_plg._obj.vue_goto_reference_spectral_value()
        assert_allclose(slice_plg.value, slice_values[0])

        slice_plg.value = slice_values[1]
        extract_plg._obj.vue_adopt_slice_as_reference()
        extract_plg._obj.vue_goto_reference_spectral_value()
        assert_allclose(slice_plg.value, slice_values[1])


@pytest.mark.parametrize('subset', ['Subset 1', 'Subset 2'])
@pytest.mark.parametrize(
    ('aperture_method', 'expected_flux_1000', 'expected_flux_2400'),
    [('Exact',
      [16.51429064, 16.52000853, 16.52572818, 16.53145005, 16.53717344, 16.54289928,
       16.54862712, 16.55435647, 16.56008781, 16.56582186],
      [26.812409, 26.821692, 26.830979, 26.840268, 26.849561, 26.858857,
       26.868156, 26.877459, 26.886765, 26.896074]),
     ('Center', 21, 25)]
)
def test_cone_aperture_with_different_methods(cubeviz_helper, spectrum1d_cube_largest,
                                              subset, aperture_method, expected_flux_1000,
                                              expected_flux_2400):
    cubeviz_helper.load_data(spectrum1d_cube_largest)
    center = PixCoord(5, 10)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        CirclePixelRegion(center, radius=2.5), combination_mode='new')
    cubeviz_helper.plugins['Subset Tools'].import_region(
        EllipsePixelRegion(center, width=5, height=5), combination_mode='new')

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = subset
    extract_plg.aperture_method.selected = aperture_method
    extract_plg.wavelength_dependent = True
    extract_plg.function = 'Sum'

    collapsed_spec = extract_plg.extract()

    assert_allclose(collapsed_spec.flux.value[1000:1010], expected_flux_1000, rtol=1e-6)
    assert_allclose(collapsed_spec.flux.value[2400:2410], expected_flux_2400, rtol=1e-6)

    extract_plg.function = 'Mean'
    collapsed_spec_mean = extract_plg.extract()

    assert_allclose(collapsed_spec_mean.flux.value, 1)


@pytest.mark.parametrize('subset', ['Subset 1', 'Subset 2'])
@pytest.mark.parametrize(
    ('aperture_method', 'expected_flux_wav'),
    [('Exact', 19.6349540849),
     ('Center', 21)]
)
def test_cylindrical_aperture_with_different_methods(cubeviz_helper, spectrum1d_cube_largest,
                                                     subset, aperture_method, expected_flux_wav):
    cubeviz_helper.load_data(spectrum1d_cube_largest, data_label="test")
    center = PixCoord(5, 10)
    cubeviz_helper.plugins['Subset Tools'].import_region([
        CirclePixelRegion(center, radius=2.5),
        EllipsePixelRegion(center, width=5, height=5)], combination_mode='new')

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = subset
    extract_plg.aperture_method.selected = aperture_method
    extract_plg.wavelength_dependent = False
    extract_plg.function = 'Sum'

    collapsed_spec = extract_plg.extract()

    assert_allclose(collapsed_spec.flux.value, expected_flux_wav)

    extract_plg.function = 'Mean'
    collapsed_spec_mean = extract_plg.extract()

    assert_allclose(collapsed_spec_mean.flux.value, 1)


# NOTE: Not as thorough as circle and ellipse above but good enough.
def test_rectangle_aperture_with_exact(cubeviz_helper, spectrum1d_cube_largest):
    cubeviz_helper.load_data(spectrum1d_cube_largest)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        RectanglePixelRegion(PixCoord(5, 10), width=4, height=4))

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = "Subset 1"
    extract_plg.aperture_method.selected = "Exact"
    extract_plg.wavelength_dependent = True
    extract_plg.function = 'Sum'
    collapsed_spec = extract_plg.extract()

    # The extracted spectrum has "steps" (aliased) but perhaps that is due to
    # how photutils is extracting a boxy aperture. There is still a slope.
    expected_flux_step = [9.378906, 10.5625, 11.816406, 13.140625, 14.535156,
                          16, 17.535156, 19.691406, 21.972656, 24.378906]
    assert_allclose(collapsed_spec.flux.value[::301], expected_flux_step)

    extract_plg.wavelength_dependent = False
    collapsed_spec = extract_plg.extract()

    assert_allclose(collapsed_spec.flux.value, 16)  # 4 x 4


def test_background_subtraction(cubeviz_helper, spectrum1d_cube_largest):
    # add constant background:
    spectrum1d_cube_largest = spectrum1d_cube_largest + 1 * u.Jy

    cubeviz_helper.load_data(spectrum1d_cube_largest)
    cubeviz_helper.plugins['Subset Tools'].import_region([
        CirclePixelRegion(PixCoord(5, 10), radius=2.5),
        EllipsePixelRegion(PixCoord(13, 10), width=3, height=5)], combination_mode='new')

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']
    with extract_plg.as_active():
        extract_plg.aperture = 'Subset 1'
        spec_no_bg = extract_plg.extract()

        extract_plg.background = 'Subset 2'

        # test visiblity of background aperture and preview based on "active step"
        assert extract_plg.background.marks[0].visible
        assert not extract_plg._obj.marks['bg_extract'].visible
        extract_plg._obj.active_step = 'ap'
        assert not extract_plg.background.marks[0].visible
        assert not extract_plg._obj.marks['bg_extract'].visible
        extract_plg._obj.active_step = 'bg'
        assert extract_plg.background.marks[0].visible
        assert extract_plg._obj.marks['bg_extract'].visible

        bg_spec = extract_plg.extract_bg_spectrum()
        extract_plg.bg_spec_per_spaxel = True
        bg_spec_normed = extract_plg.extract_bg_spectrum()
        assert np.all(bg_spec_normed.flux.value < bg_spec.flux.value)
        spec = extract_plg.extract()

    assert np.allclose(spec.flux, spec_no_bg.flux - bg_spec.flux)

    # number of pixels in the aperture, background subsets:
    n_aperture_pixels = cubeviz_helper.app.get_subsets('Subset 1')[0]['region'].to_mask().data.sum()
    n_bg_pixels = cubeviz_helper.app.get_subsets('Subset 2')[0]['region'].to_mask().data.sum()

    # the background subtracted from each slice in wavelength from the aperture should be equal
    # to the background -- which is the minimum per spectral slice in this example cube -- divided
    # by the number of pixels in the aperture:
    cube_min_per_slice = extract_plg._obj.cube['flux'].min(axis=(0, 1))
    np.testing.assert_allclose(
        bg_spec.flux.value / n_aperture_pixels,
        cube_min_per_slice
    )

    # background normalized per spaxel should be equal to the minimum per spectral slice:
    cube_min_per_slice = extract_plg._obj.cube['flux'].min(axis=(0, 1))
    np.testing.assert_allclose(
        bg_spec_normed.flux.value / n_bg_pixels,
        cube_min_per_slice
    )


def test_cone_and_cylinder_errors(cubeviz_helper, spectrum1d_cube_largest):
    cubeviz_helper.load_data(spectrum1d_cube_largest)
    center = PixCoord(5, 10)
    cubeviz_helper.plugins['Subset Tools'].import_region([
        CirclePixelRegion(center, radius=2.5),
        CircleAnnulusPixelRegion(center, inner_radius=2.5, outer_radius=4)], combination_mode='new')

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = 'Subset 1'
    extract_plg.aperture_method.selected = 'Exact'
    extract_plg.wavelength_dependent = True

    extract_plg.function = 'Min'
    with pytest.raises(ValueError, match=extract_plg._obj.conflicting_aperture_error_message):
        extract_plg.extract()

    extract_plg.function = 'Max'
    with pytest.raises(ValueError, match=extract_plg._obj.conflicting_aperture_error_message):
        extract_plg.extract()

    extract_plg.function = 'Sum'
    extract_plg.aperture = 'Subset 2'
    with pytest.raises(NotImplementedError, match=".* is not supported"):
        extract_plg.extract()


def test_cone_aperture_with_frequency_units(cubeviz_helper, spectral_cube_wcs):
    data = Spectrum1D(flux=np.ones((128, 129, 256)) * u.nJy, wcs=spectral_cube_wcs)
    cubeviz_helper.load_data(data, data_label="Test Flux")
    cubeviz_helper.plugins['Subset Tools'].import_region(
        [CirclePixelRegion(PixCoord(14, 15), radius=2.5)])

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']

    extract_plg.aperture = 'Subset 1'
    extract_plg.aperture_method.selected = 'Exact'
    extract_plg.wavelength_dependent = True
    extract_plg.function = 'Sum'

    with pytest.raises(ValueError, match="Spectral axis unit physical type is"):
        extract_plg.extract()


def test_cube_extraction_with_nan(cubeviz_helper, image_cube_hdu_obj):
    image_cube_hdu_obj[1].data[:, :2, :2] = np.nan
    cubeviz_helper.load_data(image_cube_hdu_obj, data_label="with_nan")
    extract_plg = cubeviz_helper.plugins['Spectral Extraction']
    sp = extract_plg.extract()  # Default settings (sum)
    assert_allclose(sp.flux.value, 9.6E-16)  # (10 x 10) - 4

    cubeviz_helper.plugins['Subset Tools'].import_region(
        RectanglePixelRegion(PixCoord(1.5, 1.5), width=4, height=4))
    extract_plg.aperture = 'Subset 1'
    sp_subset = extract_plg.extract()  # Default settings but on Subset
    assert_allclose(sp_subset.flux.value, 1.2E-16)  # (4 x 4) - 4


def test_autoupdate_results(cubeviz_helper, spectrum1d_cube_largest):
    cubeviz_helper.load_data(spectrum1d_cube_largest)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        CircularROI(xc=5, yc=5, radius=2))

    extract_plg = cubeviz_helper.plugins['Spectral Extraction']
    extract_plg.aperture = 'Subset 1'
    extract_plg.add_results.label = 'extracted'
    extract_plg.add_results._obj.auto_update_result = True
    _ = extract_plg.extract()

#    orig_med_flux = np.median(cubeviz_helper.get_data('extracted').flux)

    # replace Subset 1 with a larger subset, resulting fluxes should increase
    cubeviz_helper.plugins['Subset Tools'].combination_mode = 'replace'
    cubeviz_helper.plugins['Subset Tools'].import_region(CircularROI(xc=5, yc=5, radius=3))

    # update should take place automatically, but since its async, we'll call manually to ensure
    # the update is complete before comparing results
    for subset in cubeviz_helper.app.data_collection.subset_groups[0].subsets:
        cubeviz_helper.app._update_live_plugin_results(trigger_subset=subset)
    # TODO: this is randomly failing in CI (not always) so will disable the assert for now and just
    # cover to make sure the logic does not crash
#    new_med_flux = np.median(cubeviz_helper.get_data('extracted').flux)
#    assert new_med_flux > orig_med_flux


def test_aperture_composite_detection(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube)
    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    spec_extr_plugin = cubeviz_helper.plugins['Spectral Extraction']._obj

    # create a rectangular subset with all spaxels:
    rectangle = RectangularROI(-0.5, 1.5, -0.5, 3.5)
    subset_plugin.import_region(rectangle)

    # select subset 1, ensure it's not a composite subset:
    spec_extr_plugin.aperture_selected = 'Subset 1'
    assert not spec_extr_plugin.aperture.is_composite

    # now remove from this subset a circular region in the center:
    circle = CircularROI(0.5, 1.5, 1)

    subset_plugin.combination_mode = 'andnot'
    subset_plugin.import_region(circle)

    # now the subset is composite:
    assert spec_extr_plugin.aperture.is_composite


def test_extraction_composite_subset(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube)

    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    spec_extr_plugin = cubeviz_helper.plugins['Spectral Extraction']._obj

    lower_aperture = RectangularROI(-0.5, 0.5, -0.5, 1.5)
    upper_aperture = RectangularROI(2.5, 3.5, -0.5, 1.5)

    subset_plugin.import_region(lower_aperture)
    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(upper_aperture)

    spec_extr_plugin.aperture_selected = 'Subset 1'
    spectrum_1 = spec_extr_plugin.extract()

    spec_extr_plugin.aperture_selected = 'Subset 2'
    spectrum_2 = spec_extr_plugin.extract()

    rectangle = RectangularROI(-0.5, 3.5, -0.5, 1.5)

    subset_plugin.combination_mode = 'new'
    subset_plugin.import_region(rectangle)

    subset_plugin._obj.subset_selected = 'Subset 3'
    circle = CircularROI(1.5, 0.5, 1.1)

    subset_plugin.combination_mode = 'andnot'
    subset_plugin.import_region(circle)

    spec_extr_plugin.aperture_selected = 'Subset 3'

    assert spec_extr_plugin.aperture.is_composite

    spectrum_3 = spec_extr_plugin.extract()

    np.testing.assert_allclose(
        (spectrum_1 + spectrum_2).flux.value,
        (spectrum_3).flux.value
    )


def test_spectral_extraction_with_correct_sum_units(cubeviz_helper,
                                                    spectrum1d_cube_fluxunit_jy_per_steradian):
    cubeviz_helper.load_data(spectrum1d_cube_fluxunit_jy_per_steradian)
    spec_extr_plugin = cubeviz_helper.plugins['Spectral Extraction']._obj
    collapsed = spec_extr_plugin.extract()

    assert '_pixel_scale_factor' in collapsed.meta

    # Original units in Jy / sr
    # After collapsing, sr is removed via the scale factor and the extracted spectrum is in Jy
    expected_flux_values = (np.array([190., 590., 990., 1390., 1790.,
                                      2190., 2590., 2990., 3390., 3790.]) *
                            collapsed.meta.get('_pixel_scale_factor'))

    np.testing.assert_allclose(
        collapsed.flux.value,
        expected_flux_values
    )
    assert collapsed.flux.unit == u.Jy
    assert collapsed.uncertainty.unit == u.Jy


def test_default_spectral_extraction(cubeviz_helper, spectrum1d_cube_fluxunit_jy_per_steradian):
    # spacetelescope/jdaviz#3086 reported that the default cube
    # spectral extraction in cubeviz did not match the spectral extraction
    # for a spatial subset that captures all data-containing spaxels. this
    # regression tests make sure that doesn't happen anymore by accounting
    # for non-science pixels in the sums:
    cubeviz_helper.load_data(spectrum1d_cube_fluxunit_jy_per_steradian)

    subset_plugin = cubeviz_helper.plugins['Subset Tools']

    subset_plugin.import_region(CircularROI(1.5, 2, 5))

    # the first and second spectra correspond to the default extraction
    # and the subset extraction. the fluxes in these extractions should agree:
    extracted_spectra = list(
        cubeviz_helper.specviz.get_spectra(apply_slider_redshift=False).values()
    )

    assert_quantity_allclose(
        extracted_spectra[0].flux, extracted_spectra[1].flux
    )


def test_spectral_extraction_unit_conv_one_spec(
    cubeviz_helper, spectrum1d_cube_fluxunit_jy_per_steradian
):
    cubeviz_helper.load_data(spectrum1d_cube_fluxunit_jy_per_steradian)
    spectrum_viewer = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_spectrum_viewer_reference_name)
    uc = cubeviz_helper.plugins["Unit Conversion"]
    assert uc.flux_unit == "Jy"
    uc.flux_unit.selected = "MJy"
    assert spectrum_viewer.state.y_display_unit == "MJy"
    spec_extr_plugin = cubeviz_helper.plugins['Spectral Extraction']
    # Overwrite the one and only default extraction.
    collapsed = spec_extr_plugin.extract()
    # Actual values not in display unit but should not affect display unit.
    assert collapsed.flux.unit == u.Jy
    assert uc.flux_unit.selected == "MJy"
    assert spectrum_viewer.state.y_display_unit == "MJy"


@pytest.mark.usefixtures('_jail')
@pytest.mark.remote_data
@pytest.mark.parametrize(
    "start_slice, aperture, expected_rtol, uri, calspec_url",
    (
        (5.2, (20.5, 17, 10.9), 0.03,
         "mast:jwst/product/jw01524-o003_t002_miri_ch1-shortmediumlong_s3d.fits",
         calspec_url + "delumi_mod_005.fits"),  # delta UMi

        (4.85, (28, 21, 12), 0.03,
         "mast:jwst/product/jw01050-o003_t005_miri_ch1-shortmediumlong_s3d.fits",
         calspec_url + "hd159222_mod_007.fits"),  # HD 159222
    )
)
def test_spectral_extraction_scientific_validation(
    cubeviz_helper, start_slice,
    aperture, expected_rtol, uri, calspec_url
):
    """
    Compare the extracted spectrum from MIRI CH1 IFU cubes for
    delta UMi (A1Van star) and HD 159222 (G1V star) with CALSPEC model
    spectra. These model spectra are already flux calibrated using other
    observatories, and are used to compute the flux calibration for JWST.
    They are available for download at:
    https://archive.stsci.edu/hlsps/reference-atlases/cdbs/current_calspec/

    We can compare the spectral extraction results from Cubeviz with these model
    spectra, without further normalization of the models or data. When comparing
    any flux-calibrated observation against any model, the expected mismatch can be
    as large as ~10%. The agreement is much better for stars used to compute
    the flux calibration, which are the targets selected for this test.

    Both observations in this test are from MIRI. It's likely that the
    accuracy of the flux calibration of the data products available on MAST will
    evolve with time. For the latest updates on MIRI flux calibration, see:
    https://jwst-docs.stsci.edu/jwst-calibration-status/miri-calibration-status/
    """
    # Download CALSPEC model spectrum, initialize Spectrum1D.
    calspec_fitsrec = fits.getdata(calspec_url)
    column_units = [u.AA] + 2 * [u.Unit('erg s-1 cm-2 AA-1')]
    spectra_table = QTable(calspec_fitsrec, units=column_units)
    model_spectrum = Spectrum1D(
        flux=spectra_table['FLUX'],
        spectral_axis=spectra_table['WAVELENGTH']
    )

    # load observations into Cubeviz
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        cubeviz_helper.load_data(uri, cache=True)

    # add a subset with an aperture centered on each source
    subset_plugin = cubeviz_helper.plugins['Subset Tools']
    subset_plugin.import_region(CircularROI(*aperture))

    # set the slice to the blue end of MIRI CH1
    slice_plugin = cubeviz_helper.plugins['Slice']
    slice_plugin.value = start_slice

    # run a conical spectral extraction
    spectral_extraction = cubeviz_helper.plugins['Spectral Extraction']
    spectral_extraction.aperture = 'Subset 1'
    spectral_extraction.wavelength_dependent = True
    spectral_extraction._obj.results_label = 'conical-extraction'
    extracted_spectrum = spectral_extraction.extract()

    # resample the model spectrum on the same wavelengths as
    # the observed spectrum:
    resampled_spectrum = FluxConservingResampler()(
        model_spectrum, extracted_spectrum.wavelength
    )

    # load model spectrum:
    cubeviz_helper.specviz.load_data(resampled_spectrum, data_label='calspec model')

    # compute the relative residual, take the median absolute deviation:
    median_abs_relative_dev = abs(np.median(
        abs(
            extracted_spectrum.flux /
            resampled_spectrum.flux.to(u.MJy, u.spectral_density(resampled_spectrum.wavelength))
        ).to_value(u.dimensionless_unscaled) - 1
    ))
    assert median_abs_relative_dev < expected_rtol


@pytest.mark.parametrize("flux_angle_unit", [(u.Unit(x), u.sr) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS]  # noqa
                                              + [(u.Unit(x), PIX2) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS])  # noqa
def test_spectral_extraction_flux_unit_conversions(cubeviz_helper,
                                                   spectrum1d_cube_custom_fluxunit,
                                                   flux_angle_unit):
    """
    Test that cubeviz spectral extraction plugin works with all possible
    flux unit conversions for a cube loaded in units spectral/photon flux
    density. The spectral extraction result will remain in the native
    data unit, but the extraction preview should be converted to the
    display unit.
    """

    flux_unit, angle_unit = flux_angle_unit

    sb_cube = spectrum1d_cube_custom_fluxunit(fluxunit=flux_unit / angle_unit,
                                              shape=(5, 4, 4),
                                              with_uncerts=True)
    cubeviz_helper.load_data(sb_cube)

    spectrum_viewer = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_spectrum_viewer_reference_name)

    uc = cubeviz_helper.plugins["Unit Conversion"]
    se = cubeviz_helper.plugins['Spectral Extraction']
    se.keep_active = True  # keep active for access to preview markers

    # equivalencies for unit conversion, for comparison of outputs
    equivs = all_flux_unit_conversion_equivs(se.dataset.selected_obj.meta.get('PIXAR_SR', 1.0),
                                             se.dataset.selected_obj.spectral_axis)

    for new_flux_unit in SPEC_PHOTON_FLUX_DENSITY_UNITS:
        if new_flux_unit != flux_unit:

            uc.flux_unit.selected = flux_unit.to_string()
            uc.spectral_y_type.selected = 'Flux'

            # and set back to sum initially so units will always be flux not sb
            se.function.selected = 'Sum'
            se.extract()

            original_sum_y_values = se._obj.marks['extract'].y

            # set to new unit
            uc.flux_unit.selected = new_flux_unit
            assert spectrum_viewer.state.y_display_unit == new_flux_unit

            # still using 'sum', results should be in flux
            collapsed = se.extract()

            # make sure extraction preview was translated to new display units
            new_sum_y_values = se._obj.marks['extract'].y
            new_converted_to_old_unit = flux_conversion_general(new_sum_y_values,
                                                                u.Unit(new_flux_unit),
                                                                u.Unit(flux_unit),
                                                                equivs, with_unit=False)
            np.testing.assert_allclose(original_sum_y_values, new_converted_to_old_unit)

            # collapsed result will still have the native data flux unit
            assert uc.spectral_y_type.selected == 'Flux'
            assert collapsed.flux.unit == collapsed.uncertainty.unit == flux_unit
            # but display units in spectrum viewer should reflect new flux unit selection
            assert se._obj.spectrum_y_units == se._obj.results_units == new_flux_unit



import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.nddata import NDData
from numpy.testing import assert_allclose

from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS, create_example_gwcs


class TestPanZoomTools(BaseImviz_WCS_WCS):
    def test_panzoom_tools(self):
        v = self.imviz.default_viewer._obj
        v2 = self.imviz.create_image_viewer()
        self.imviz.app.add_data_to_viewer('imviz-1', 'has_wcs_1[SCI,1]')

        t = v.toolbar.tools['jdaviz:boxzoommatch']
        # original limits (x_min, x_max, y_min, y_max): -0.5 9.5 -0.5 9.5
        # original limits (zoom_center_x, zoom_center_y, zoom_radius): 4.5 4.5 5.0
        original_limits = v.get_limits()
        assert_allclose(original_limits, (-0.5, 9.5, -0.5, 9.5))
        original_center_rad = (v.state.zoom_center_x, v.state.zoom_center_y, v.state.zoom_radius)
        assert_allclose(original_center_rad, (4.5, 4.5, 5.0))
        assert_allclose(v2.get_limits(), original_limits)
        assert_allclose((v2.state.zoom_center_x, v2.state.zoom_center_y, v2.state.zoom_radius), original_center_rad)  # noqa
        t.activate()
        t.save_prev_zoom()
        v.set_limits(x_min=1, x_max=8, y_min=1, y_max=8)
        # second viewer should match these changes, wrt zoom center and radius
        assert v2.state.zoom_center_x == v.state.zoom_center_x
        assert v2.state.zoom_center_y == v.state.zoom_center_y
        assert v2.state.zoom_radius == v.state.zoom_radius

        v.toolbar.tools['jdaviz:prevzoom'].activate()
        # both should revert since they're still linked (boxzoommatch will re-activate)
        assert_allclose((v.state.zoom_center_x, v.state.zoom_center_y, v.state.zoom_radius), original_center_rad)  # noqa
        assert_allclose((v2.state.zoom_center_x, v2.state.zoom_center_y, v2.state.zoom_radius), original_center_rad)  # noqa

        v.toolbar.tools['jdaviz:prevzoom'].activate()
        # both should revert since they're still linked
        assert v.get_limits() == (1, 8, 1, 8)
        assert v2.get_limits() == (1, 8, 1, 8)

        v.toolbar.tools['jdaviz:boxzoommatch'].deactivate()
        v.toolbar.tools['jdaviz:homezoom'].activate()
        assert_allclose(v.get_limits(), original_limits)
        assert_allclose(v2.get_limits(), (1, 8, 1, 8))
        v.toolbar.tools['jdaviz:prevzoom'].activate()
        assert_allclose(v.get_limits(), (1, 8, 1, 8))
        assert_allclose(v2.get_limits(), (1, 8, 1, 8))
        t.deactivate()

        t_linkedpan = v.toolbar.tools['jdaviz:panzoommatch']
        t_linkedpan.activate()
        v.center_on((0, 0))
        # make sure both viewers moved to the new center
        assert_allclose(v.get_limits(), (-3.5, 3.5, -3.5, 3.5))
        assert_allclose(v2.get_limits(), (-3.5, 3.5, -3.5, 3.5))
        t_linkedpan.deactivate()

        t_normpan = v.toolbar.tools['jdaviz:imagepanzoom']
        t_normpan.activate()
        t_normpan.on_click({'event': 'click', 'domain': {'x': 1, 'y': 1}})
        # make sure only first viewer re-centered since this mode is not linked mode
        assert_allclose(v.get_limits(), (-2.5, 4.5, -2.5, 4.5))
        assert_allclose(v2.get_limits(), (-3.5, 3.5, -3.5, 3.5))
        t_normpan.deactivate()

        t_linkedpan.activate()
        t_linkedpan.on_click({'event': 'click', 'domain': {'x': 2, 'y': 2}})
        # make sure both viewers moved to the new center
        assert_allclose(v.get_limits(), (-1.5, 5.5, -1.5, 5.5))
        assert_allclose(v2.get_limits(), (-1.5, 5.5, -1.5, 5.5))
        t_linkedpan.deactivate()


@pytest.mark.parametrize("align_by", ["Pixels", "WCS"])
def test_panzoom_click_center_linking(imviz_helper, align_by):
    """https://github.com/spacetelescope/jdaviz/issues/2749"""
    v = imviz_helper.default_viewer._obj

    # Since we are not really displaying, need this to test pan/zoom.
    v.shape = (100, 100)
    v.state._set_axes_aspect_ratio(1)

    arr_big = np.ones((40, 30), dtype=int)
    w_big = create_example_gwcs(arr_big.shape)
    arr_small = np.ones((20, 15), dtype=int)
    w_small = create_example_gwcs(arr_small.shape)

    imviz_helper.load_data(NDData(arr_big, wcs=w_big), data_label="big")
    imviz_helper.load_data(NDData(arr_small, wcs=w_small), data_label="small")

    lc_plugin = imviz_helper.plugins['Orientation']
    lc_plugin.align_by = align_by

    coo = SkyCoord(ra=197.89262754541807, dec=-1.3644568140486624, unit="deg")

    if align_by == "WCS":
        mouseover_loc = v.state.reference_data.coords.world_to_pixel(coo)
    else:  # Pixels
        mouseover_loc = w_small.world_to_pixel(coo)

    t = v.toolbar.tools['jdaviz:imagepanzoom']
    t.activate()
    t.on_click({'event': 'click', 'domain': {'x': mouseover_loc[0], 'y': mouseover_loc[1]}})
    t.deactivate()

    # We want to make sure click centers viewer to where it is supposed to be.
    cur_cen = v._get_center_skycoord()
    v.center_on(coo)
    real_cen = v._get_center_skycoord()
    assert_allclose(cur_cen.ra.deg, real_cen.ra.deg)
    assert_allclose(cur_cen.dec.deg, real_cen.dec.deg)


def test_blink(imviz_helper):
    viewer = imviz_helper.default_viewer._obj

    for i in range(3):
        imviz_helper.load_data(np.zeros((2, 2)) + i, data_label=f'image_{i}')

    label_mouseover = imviz_helper.app.session.application._tools['g-coords-info']
    viewer.on_mouse_or_key_event({'event': 'keydown', 'key': 'b', 'domain': {'x': 0, 'y': 0}})
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove', 'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +0.00000e+00', '', '')
    assert viewer.top_visible_data_label == 'image_0'

    # Blink forward and update coordinates info panel.
    viewer.blink_once()
    assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +1.00000e+00', '', '')
    assert viewer.top_visible_data_label == 'image_1'

    # Blink backward.
    viewer.blink_once(reversed=True)
    assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +0.00000e+00', '', '')
    assert viewer.top_visible_data_label == 'image_0'

    # Blink backward again.
    viewer.on_mouse_or_key_event({'event': 'keydown', 'key': 'B', 'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == ('Pixel x=00.0 y=00.0 Value +2.00000e+00', '', '')
    assert viewer.top_visible_data_label == 'image_2'


def test_compass_open_while_load(imviz_helper):
    plg = imviz_helper.plugins['Compass']
    plg._obj.plugin_opened = True

    # Should not crash even if Compass is open in tray.
    imviz_helper.load_data(np.ones((2, 2)))
    assert len(imviz_helper.app.data_collection) == 1


def test_tool_visibility(imviz_helper):
    imviz_helper.load_data(np.ones((2, 2)))
    tb = imviz_helper.default_viewer._obj.toolbar

    assert not tb.tools_data['jdaviz:boxzoommatch']['visible']

    assert tb.tools_data['jdaviz:boxzoom']['primary']
    # activate boxzoom to ensure it remains primary
    tb.active_tool_id = 'jdaviz:boxzoom'

    imviz_helper.create_image_viewer()
    imviz_helper.app.set_data_visibility('imviz-1', imviz_helper.app.data_collection[0].label, True)

    assert tb.tools_data['jdaviz:boxzoommatch']['visible']
    assert tb.active_tool_id == 'jdaviz:boxzoom'
    assert tb.tools_data['jdaviz:boxzoom']['primary']

    # but the panzoom has updated primary since there was no active tool in that submenu
    assert tb.tools_data['jdaviz:panzoommatch']['visible']
    assert tb.tools_data['jdaviz:panzoommatch']['primary']

    # now set the tool to the matched box zoom to ensure it deactivates itself when removing
    # a viewer
    tb.active_tool_id = 'jdaviz:boxzoommatch'
    imviz_helper.destroy_viewer('imviz-1')
    assert not tb.tools_data['jdaviz:boxzoommatch']['visible']
    assert tb.active_tool_id is None



import pytest

from jdaviz import __version__
from jdaviz.configs.default.plugins.about.about import latest_version_from_pypi


@pytest.mark.remote_data
def test_get_latest_version_from_pypi():
    # Since the latest version will change over time,
    # we can only check that something is returned but not what exactly.

    v = latest_version_from_pypi("jdaviz")
    assert isinstance(v, str)

    v = latest_version_from_pypi("fakespacetelescopepackage")
    assert v is None


# NOTE: Theoretically now all the tests in jdaviz are remote data tests
# because all the viz loads About plugin and About plugin always polls
# PyPI for the latest release version. If that bothers you or PyPI
# starts blocking us as spammer, we can consider caching it at package level
# but that introduces new non-standard package attributes.
def test_about_basic(specviz_helper):
    plg = specviz_helper.plugins["About"]._obj

    assert plg.jdaviz_version == __version__
    assert isinstance(plg.jdaviz_pypi, str)
    # not_is_latest can be non-deterministic because user can be running
    # this test from an older version going forward, so we skip checking it.



import os

import pytest
from astropy.wcs import FITSFixedWarning

from jdaviz.utils import alpha_index, download_uri_to_path


@pytest.mark.parametrize("test_input,expected", [(0, 'a'), (1, 'b'), (25, 'z'), (26, 'aa'),
                                                 (701, 'zz'), (702, '{a')])
def test_alpha_index(test_input, expected):
    assert alpha_index(test_input) == expected


def test_alpha_index_exceptions():
    with pytest.raises(TypeError, match="index must be an integer"):
        alpha_index(4.2)
    with pytest.raises(ValueError, match="index must be positive"):
        alpha_index(-1)


def test_uri_to_download_bad_scheme(imviz_helper):
    uri = "file://path/to/file.fits"
    with pytest.raises(ValueError, match=r'URI file://path/to/file\.fits with scheme file'):
        imviz_helper.load_data(uri)


@pytest.mark.remote_data
def test_uri_to_download_nonexistent_mast_file(imviz_helper):
    # this validates as a mast uri but doesn't actually exist on mast:
    uri = "mast:JWST/product/jw00000-no-file-here.fits"
    with pytest.raises(ValueError, match='Failed query for URI'):
        imviz_helper.load_data(uri, cache=False)


@pytest.mark.remote_data
def test_url_to_download_imviz_local_path_warning(imviz_helper):
    url = "https://www.astropy.org/astropy-data/tutorials/FITS-images/HorseHead.fits"
    match_local_path_msg = (
        'You requested to cache data to the .*local_path.*supported for downloads of '
        'MAST URIs.*astropy download cache instead.*'
    )
    with (
        pytest.warns(FITSFixedWarning, match="'datfix' made the change"),
        pytest.warns(UserWarning, match=match_local_path_msg)
    ):
        imviz_helper.load_data(url, cache=True, local_path='horsehead.fits')


def test_uri_to_download_specviz_local_path_check():
    uri = "mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits"
    local_path = download_uri_to_path(uri, cache=False, dryrun=True)  # No download

    # Wrong: '.\\JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits'
    # Correct:  '.\\jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits'
    assert local_path == os.path.join(os.curdir, "jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits")  # noqa: E501


@pytest.mark.remote_data
def test_uri_to_download_specviz(specviz_helper, tmp_path):
    uri = "mast:JWST/product/jw02732-o004_t004_miri_ch1-shortmediumlong_x1d.fits"
    local_path = str(tmp_path / uri.split('/')[-1])
    specviz_helper.load_data(uri, cache=True, local_path=local_path)


@pytest.mark.remote_data
def test_uri_to_download_specviz2d(specviz2d_helper, tmp_path):
    uri = "mast:jwst/product/jw01538-o161_s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits"
    local_path = str(tmp_path / uri.split('/')[-1])
    specviz2d_helper.load_data(uri, cache=True, local_path=local_path)



# Tests all non-data-loading functionality of the Mosviz Helper class
# For data-loading tests, see test_data_loading.py

import csv

import pytest
from numpy.testing import assert_allclose

from jdaviz.configs.specviz2d.helper import Specviz2d


def test_to_csv(tmp_path, mosviz_helper, spectrum_collection):
    labels = [f"Test Spectrum Collection {i}" for i in range(5)]
    mosviz_helper.load_1d_spectra(spectrum_collection, data_labels=labels, add_redshift_column=True)

    mosviz_helper.to_csv(filename=str(tmp_path / "MOS_data.csv"))

    found_rows = 0
    found_index_label = False

    with open(tmp_path / "MOS_data.csv", "r") as f:
        freader = csv.reader(f)
        for row in freader:
            if row[0] == "Table Index":
                found_index_label = True
            else:
                found_rows += 1

    assert found_index_label
    assert found_rows == 5


def test_table_scrolling(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d):
    spectra1d = [spectrum1d] * 2
    spectra2d = [mos_spectrum2d] * 2

    mosviz_helper.load_data(spectra1d, spectra2d, images=mos_image)

    table = mosviz_helper.app.get_viewer('table-viewer')
    # first row is automatically selected in the UI
    # (otherwise it would be None which is a case not handled)
    table.widget_table.highlighted = 0
    table.next_row()
    assert table.widget_table.highlighted == 1
    table.next_row()
    # with only 2 rows, this should wrap back to 0
    assert table.widget_table.highlighted == 0
    table.prev_row()
    assert table.widget_table.highlighted == 1


def test_column_visibility(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d):
    spectra1d = [spectrum1d] * 2
    spectra2d = [mos_spectrum2d] * 2

    mosviz_helper.load_data(spectra1d, spectra2d, images=mos_image)

    with pytest.raises(
                ValueError,
                match="visible must be one of None, True, or False."):
        mosviz_helper.get_column_names(visible='string')

    assert 'Redshift' not in mosviz_helper.get_column_names(True)
    assert 'Redshift' in mosviz_helper.get_column_names(False)
    assert 'Redshift' in mosviz_helper.get_column_names()

    mosviz_helper.show_column('Redshift')
    assert 'Redshift' in mosviz_helper.get_column_names(True)

    mosviz_helper.hide_column('Redshift')
    assert 'Redshift' not in mosviz_helper.get_column_names(True)


def test_custom_columns(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d):
    spectra1d = [spectrum1d] * 2
    spectra2d = [mos_spectrum2d] * 2

    mosviz_helper.load_data(spectra1d, spectra2d, images=mos_image)

    mosviz_helper.add_column('custom_name')
    assert 'custom_name' in mosviz_helper.get_column_names(True)
    assert len(mosviz_helper.get_column('custom_name')) == 2

    mosviz_helper.update_column('custom_name', 0.1, row=1)
    assert mosviz_helper.get_column('custom_name')[1] == 0.1

    with pytest.raises(
                ValueError,
                match="row out of range of table"):
        mosviz_helper.update_column('custom_name', 0.3, row=3)

    with pytest.raises(
                ValueError,
                match="data must have length 2 \\(rows in table\\)"):
        mosviz_helper.add_column('custom_name_2', [0.1])

    mosviz_helper.show_column("Redshift")
    assert "Redshift" in mosviz_helper.get_column_names(True)
    mosviz_helper.set_visible_columns()
    assert "Redshift" not in mosviz_helper.get_column_names(True)


def test_redshift_column(mosviz_helper, mos_image, spectrum1d, mos_spectrum2d):
    spectra1d = [spectrum1d] * 2
    spectra2d = [mos_spectrum2d] * 2

    mosviz_helper.load_data(spectra1d, spectra2d, images=mos_image)

    mosviz_helper.update_column("Redshift", 0.1, row=0)
    all_spectra = mosviz_helper.specviz.get_spectra(apply_slider_redshift=True)
    assert_allclose(list(all_spectra.values())[0].redshift.value, 0.1)
    assert isinstance(mosviz_helper.specviz2d, Specviz2d)
    assert_allclose(mosviz_helper.get_spectrum_1d(apply_slider_redshift=True).redshift.value, 0.1)
    assert_allclose(mosviz_helper.get_spectrum_2d(apply_slider_redshift=True).redshift.value, 0.1)
    assert_allclose(mosviz_helper.get_spectrum_1d(apply_slider_redshift=True, row=1).redshift.value, 0.0)  # noqa: E501


# Some API might be going through deprecation, so ignore the warning.
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_plugin_user_apis(mosviz_helper):
    for plugin_name, plugin_api in mosviz_helper.plugins.items():
        plugin = plugin_api._obj
        for attr in plugin_api._expose:
            assert hasattr(plugin, attr)



import json

import requests
from packaging.version import Version
from traitlets import Bool, Unicode

from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin

try:
    from jdaviz import __version__
except ImportError:  # pragma: no cover
    __version__ = "unknown"

__all__ = ['About']


@tray_registry('about', label="About")
class About(PluginTemplateMixin):
    """Show information about Jdaviz."""
    template_file = __file__, "about.vue"

    jdaviz_version = Unicode("unknown").tag(sync=True)
    jdaviz_pypi = Unicode("unknown").tag(sync=True)
    not_is_latest = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jdaviz_version = __version__

        # description displayed under plugin title in tray
        self._plugin_description = 'Information about Jdaviz.'

        if __version__ != "unknown":
            _ver_pypi = latest_version_from_pypi("jdaviz")
            if _ver_pypi:
                self.jdaviz_pypi = _ver_pypi
                self.not_is_latest = Version(__version__) < Version(_ver_pypi)
            else:  # pragma: no cover
                self.jdaviz_pypi = "unknown"
                self.not_is_latest = False


def latest_version_from_pypi(package_name):
    """Version info for given package or `None`."""
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        r = requests.get(url, timeout=60)
    except Exception:  # nosec # pragma: no cover
        pass
    else:
        if r.ok:
            try:
                d = json.loads(r.text)
                v = d["info"]["version"]
            except Exception:  # nosec # pragma: no cover
                pass
            else:
                return v



from PyInstaller.utils.hooks import collect_data_files, copy_metadata, collect_submodules

hiddenimports = collect_submodules("numpy")
datas = collect_data_files('numpy')
datas += copy_metadata('numpy')



from functools import cached_property

import numpy as np
import astropy
from astropy import units as u
from astropy.nddata import NDDataArray, StdDevUncertainty
from traitlets import Any, Bool, Dict, Float, List, Unicode, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import SnackbarMessage, SliceValueUpdatedMessage, GlobalDisplayUnitChanged
from jdaviz.core.marks import PluginLine
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin,
                                        DatasetSelectMixin,
                                        SelectPluginComponent,
                                        ApertureSubsetSelectMixin,
                                        ApertureSubsetSelect,
                                        AddResults, AddResultsMixin,
                                        skip_if_not_tray_instance,
                                        skip_if_no_updates_since_last_active,
                                        with_spinner, with_temp_disable)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.unit_conversion_utils import (all_flux_unit_conversion_equivs,
                                               flux_conversion_general,
                                               check_if_unit_is_per_solid_angle)
from jdaviz.configs.cubeviz.plugins.parsers import _return_spectrum_with_correct_units
from jdaviz.configs.cubeviz.plugins.viewers import WithSliceIndicator


__all__ = ['SpectralExtraction']


@tray_registry(
    'cubeviz-spectral-extraction', label="Spectral Extraction", viewer_requirements='spectrum'
)
class SpectralExtraction(PluginTemplateMixin, ApertureSubsetSelectMixin,
                         DatasetSelectMixin, AddResultsMixin):
    """
    See the :ref:`Spectral Extraction Plugin Documentation <spectral-extraction>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``aperture`` (:class:`~jdaviz.core.template_mixin.ApertureSubsetSelect`):
      Subset to use for the spectral extraction, or ``Entire Cube``.
    * ``wavelength_dependent``:
      Whether the ``aperture`` should be considered wavelength-dependent.  The cone is defined
      to intersect ``aperture`` at ``reference_spectral_value``.
    * ``reference_spectral_value``:
      The wavelength that will be used to calculate the radius of the cone through the cube.
    * ``background`` (:class:`~jdaviz.core.template_mixin.ApertureSubsetSelect`):
      Subset to use for background subtraction, or ``None``.
    * ``bg_wavelength_dependent``:
      Whether the ``background`` aperture should be considered wavelength-dependent (requires
      ``wavelength_dependent`` to also be set to ``True``). The cone is defined
      to intersect ``background`` at ``reference_spectral_value``.
    * ```bg_spec_per_spaxel``:
        Whether to normalize the background per spaxel when calling ``extract_bg_spectrum``.
        Otherwise, the spectrum will be scaled by the ratio between the
        areas of the aperture and the background aperture. Only applicable if ``function`` is 'Sum'.
    * ``bg_spec_add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`extract_bg_spectrum`
    * ``aperture_method`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Method to use for extracting spectrum (and background, if applicable).
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`extract`
    """
    template_file = __file__, "spectral_extraction.vue"
    uses_active_status = Bool(True).tag(sync=True)
    show_live_preview = Bool(True).tag(sync=True)

    active_step = Unicode().tag(sync=True)

    resulting_product_name = Unicode("spectrum").tag(sync=True)
    do_auto_extraction = True
    # whether wavelength dependent options should be exposed to the user (in the UI)
    wavelength_dependent_available = Bool(True).tag(sync=True)
    bg_export_available = Bool(True).tag(sync=True)

    wavelength_dependent = Bool(False).tag(sync=True)
    reference_spectral_value = FloatHandleEmpty().tag(sync=True)
    slice_spectral_value = Float().tag(sync=True)

    bg_items = List([]).tag(sync=True)
    bg_selected = Any('').tag(sync=True)
    bg_selected_validity = Dict().tag(sync=True)
    bg_scale_factor = Float(1).tag(sync=True)
    bg_wavelength_dependent = Bool(False).tag(sync=True)

    bg_spec_per_spaxel = Bool(False).tag(sync=True)
    bg_spec_results_label = Unicode().tag(sync=True)
    bg_spec_results_label_default = Unicode().tag(sync=True)
    bg_spec_results_label_auto = Bool(True).tag(sync=True)
    bg_spec_results_label_invalid_msg = Unicode('').tag(sync=True)
    bg_spec_results_label_overwrite = Bool().tag(sync=True)
    bg_spec_add_to_viewer_items = List().tag(sync=True)
    bg_spec_add_to_viewer_selected = Unicode().tag(sync=True)
    bg_spec_spinner = Bool(False).tag(sync=True)

    function_items = List().tag(sync=True)
    function_selected = Unicode('Sum').tag(sync=True)
    filename = Unicode().tag(sync=True)
    extraction_available = Bool(False).tag(sync=True)

    results_units = Unicode().tag(sync=True)
    spectrum_y_units = Unicode().tag(sync=True)
    flux_units = Unicode().tag(sync=True)
    sb_units = Unicode().tag(sync=True)

    aperture_method_items = List().tag(sync=True)
    aperture_method_selected = Unicode('Center').tag(sync=True)

    conflicting_aperture_and_function = Bool(False).tag(sync=True)
    conflicting_aperture_error_message = Unicode('Aperture method Exact cannot be selected along'
                                                 ' with Min or Max.').tag(sync=True)

    # export_enabled controls whether saving to a file is enabled via the UI.  This
    # is a temporary measure to allow server-installations to disable saving server-side until
    # saving client-side is supported
    export_enabled = Bool(True).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # description displayed under plugin title in tray
        self._plugin_description = 'Extract a spectrum from a spectral cube.'

        self.extracted_spec = None

        self.dataset.filters = ['is_flux_cube']

        # TODO: in the future this could be generalized with support in SelectPluginComponent
        self.aperture._default_text = 'Entire Cube'
        self.aperture._manual_options = ['Entire Cube']
        self.aperture.items = [{"label": "Entire Cube"}]
        self.aperture._subset_selected_changed_callback = self._update_extract
        # need to reinitialize choices since we overwrote items and some subsets may already
        # exist.
        self.aperture._initialize_choices()
        self.aperture.select_default()

        self.background = ApertureSubsetSelect(self,
                                               'bg_items',
                                               'bg_selected',
                                               'bg_selected_validity',
                                               'bg_scale_factor',
                                               dataset='dataset',
                                               multiselect=None,
                                               default_text='None',
                                               subset_selected_changed_callback=self._update_extract)  # noqa

        self.bg_spec_add_results = AddResults(self, 'bg_spec_results_label',
                                              'bg_spec_results_label_default',
                                              'bg_spec_results_label_auto',
                                              'bg_spec_results_label_invalid_msg',
                                              'bg_spec_results_label_overwrite',
                                              'bg_spec_add_to_viewer_items',
                                              'bg_spec_add_to_viewer_selected')
        self.bg_spec_add_results.viewer.filters = ['is_slice_indicator_viewer']
        self.bg_spec_results_label_default = f'background-{self.resulting_product_name}'

        self.function = SelectPluginComponent(
            self,
            items='function_items',
            selected='function_selected',
            manual_options=['Mean', 'Min', 'Max', 'Sum']
        )
        self.aperture_method_manual_options = ['Exact', 'Center']
        self.aperture_method = SelectPluginComponent(
            self,
            items='aperture_method_items',
            selected='aperture_method_selected',
            manual_options=self.aperture_method_manual_options
        )
        self._set_default_results_label()
        self.add_results.viewer.filters = ['is_slice_indicator_viewer']

        self.session.hub.subscribe(self, SliceValueUpdatedMessage,
                                   handler=self._on_slice_changed)
        self.hub.subscribe(self, GlobalDisplayUnitChanged,
                           handler=self._on_global_display_unit_changed)

        self._update_disabled_msg()

        if self.app.state.settings.get('server_is_remote', False):
            # when the server is remote, saving the file in python would save on the server, not
            # on the user's machine, so export support in cubeviz should be disabled
            self.export_enabled = False

    @property
    def user_api(self):
        expose = ['show_live_preview', 'dataset', 'function', 'aperture',
                  'background', 'bg_wavelength_dependent',
                  'bg_spec_per_spaxel', 'bg_spec_add_results', 'extract_bg_spectrum',
                  'add_results', 'extract',
                  'wavelength_dependent', 'reference_spectral_value',
                  'aperture_method']

        return PluginUserApi(self, expose=expose)

    @observe('dataset_items')
    def _update_disabled_msg(self, msg={}):
        for data in self.app.data_collection:
            if data.data.ndim == 3:
                self.disabled_msg = ''
                break
        else:
            # no cube-like data loaded.  Once loaded, the parser will unset this
            self.disabled_msg = (
                f"{self.__class__.__name__} requires a 3d cube dataset to be loaded, "
                "please load data to enable this plugin."
            )

    @property
    def live_update_subscriptions(self):
        return {'data': ('dataset',), 'subset': ('aperture', 'background')}

    def __call__(self, add_data=True):
        return self.extract(add_data=add_data)

    @property
    def slice_display_unit_name(self):
        return 'spectral'

    @property
    def spatial_axes(self):
        # Collapse an e.g. 3D spectral cube to 1D spectrum, assuming that last axis
        # is always wavelength. This may need adjustment after the following
        # specutils PR is merged: https://github.com/astropy/specutils/pull/1033
        return (0, 1)

    @property
    def slice_indicator_viewers(self):
        return [v for v in self.app._viewer_store.values() if isinstance(v, WithSliceIndicator)]

    @observe('active_step', 'is_active')
    def _active_step_changed(self, *args):
        self.aperture._set_mark_visiblities(self.active_step in ('', 'ap', 'extract'))
        self.background._set_mark_visiblities(self.active_step == 'bg')
        self.marks['bg_extract'].visible = self.active_step == 'bg'

    @property
    def slice_plugin(self):
        return self.app._jdaviz_helper.plugins['Slice']

    @observe('aperture_items')
    @skip_if_not_tray_instance()
    def _aperture_items_changed(self, msg):
        if not self.do_auto_extraction:
            return
        if not hasattr(self, 'aperture'):
            return
        orig_labels = [item['label'] for item in msg['old']]
        for item in msg['new']:
            if item['label'] not in orig_labels:
                if item.get('type') != 'spatial':
                    continue
                subset_lbl = item.get('label')
                try:
                    self._extract_in_new_instance(subset_lbl=subset_lbl,
                                                  auto_update=True, add_data=True)
                except Exception:
                    msg = SnackbarMessage(
                        f"Automatic {self.resulting_product_name} extraction for {subset_lbl} failed",  # noqa
                        color='error', sender=self, timeout=10000)
                else:
                    msg = SnackbarMessage(
                        f"Automatic {self.resulting_product_name} extraction for {subset_lbl} successful",  # noqa
                        color='success', sender=self)
                self.app.hub.broadcast(msg)

    def _extract_in_new_instance(self, dataset=None, function='Sum', subset_lbl=None,
                                 auto_update=False, add_data=False):
        # create a new instance of the Spectral Extraction plugin (to not affect the instance in
        # the tray) and extract the entire cube with defaults.
        plg = self.new()
        plg.dataset.selected = self.dataset.selected
        if subset_lbl is not None:
            plg.aperture.selected = subset_lbl
        plg.aperture_method.selected = 'Center'
        plg.function.selected = function
        plg.add_results.auto_update_result = auto_update
        # all other settings remain at their plugin defaults
        return plg(add_data=add_data)

    @observe('wavelength_dependent', 'bg_wavelength_dependent')
    def _wavelength_dependent_changed(self, *args):
        if self.wavelength_dependent:
            self.reference_spectral_value = self.slice_plugin.value
        else:
            self.bg_wavelength_dependent = False
        # NOTE: this can be redundant in the case where reference_spectral_value changed and
        # triggers the observe, but we need to ensure it is updated if reference_spectral_value
        # is unchanged
        self._update_mark_scale()

    def _on_slice_changed(self, msg):
        self.slice_spectral_value = msg.value

    def vue_goto_reference_spectral_value(self, *args):
        self.slice_plugin.value = self.reference_spectral_value

    def vue_adopt_slice_as_reference(self, *args):
        self.reference_spectral_value = self.slice_plugin.value

    @observe('reference_spectral_value', 'slice_spectral_value')
    def _update_mark_scale(self, *args):
        if not self.wavelength_dependent:
            self.aperture.scale_factor = 1.0
        else:
            self.aperture.scale_factor = self.slice_spectral_value/self.reference_spectral_value
        if not self.bg_wavelength_dependent:
            self.background.scale_factor = 1.0
        else:
            self.background.scale_factor = self.slice_spectral_value/self.reference_spectral_value

    def _on_global_display_unit_changed(self, msg=None):
        if msg is None:
            self.flux_units = str(self.app._get_display_unit('flux'))
            self.sb_units = str(self.app._get_display_unit('sb'))
            self.spectrum_y_units = str(self.app._get_display_unit('spectral_y'))
        elif msg.axis == 'flux':
            self.flux_units = str(msg.unit)
        elif msg.axis == 'sb':
            self.sb_units = str(msg.unit)
        elif msg.axis == 'spectral_y':
            self.spectrum_y_units = str(msg.unit)
            # no need to update results_units as separate messages will have been
            # sent by unit conversion for flux and/or sb.
            # updates to spectrum_y_units will trigger updating the extraction preview
            return
        else:
            # ignore
            return
        # update results_units based on flux_units, sb_units, and currently selected function
        self._update_results_units()

    @observe('function_selected')
    def _update_results_units(self, *args):
        # NOTE this is also called by _on_global_display_unit_changed
        # after flux_units and/or sb_units is set.
        # results_units is ONLY used for the warning in the UI, so does not
        # need to trigger an update to the preview
        if self.function_selected.lower() == 'sum':
            self.results_units = self.flux_units
        else:
            self.results_units = self.sb_units

    @observe('function_selected', 'aperture_method_selected')
    def _update_aperture_method_on_function_change(self, *args):
        if (self.function_selected.lower() in ('min', 'max') and
                self.aperture_method_selected.lower() != 'center'):
            self.conflicting_aperture_and_function = True
        else:
            self.conflicting_aperture_and_function = False

    @property
    def cube(self):
        return self.dataset.selected_dc_item

    @property
    def uncert_cube(self):
        if (hasattr(self._app._jdaviz_helper, '_loaded_flux_cube') and
                hasattr(self.app._jdaviz_helper, '_loaded_uncert_cube') and
                self.dataset.selected == self._app._jdaviz_helper._loaded_flux_cube.label):
            return self._app._jdaviz_helper._loaded_uncert_cube
        else:
            # TODO: allow selecting or associating an uncertainty cube?
            return None

    @property
    def mask_cube(self):
        if (hasattr(self._app._jdaviz_helper, '_loaded_flux_cube') and
                hasattr(self.app._jdaviz_helper, '_loaded_mask_cube') and
                self.dataset.selected == self._app._jdaviz_helper._loaded_flux_cube.label):
            return self._app._jdaviz_helper._loaded_mask_cube
        else:
            # TODO: allow selecting or associating a mask/DQ cube?
            return None

    @property
    def slice_display_unit(self):
        return astropy.units.Unit(self.app._get_display_unit(self.slice_display_unit_name))

    @property
    def inverted_mask_non_science(self):
        # Aperture masks begin by removing from consideration any pixel
        # set to NaN, which corresponds to a pixel on the "non-science" portions
        # of the detector. For JWST spectral cubes, these pixels are also marked in
        # the DQ array with flag `513`. Also respect the loaded mask, if it exists.
        # This "inverted mask" is `True` where the data are included, `False` where excluded.
        mask_non_science = np.isnan(self.dataset.selected_obj.flux.value)
        if self.mask_cube is not None:
            mask_non_science = np.logical_or(self.mask_cube.get_component('flux').data,
                                             mask_non_science)
        return np.logical_not(mask_non_science).astype(float)

    @property
    def aperture_weight_mask(self):
        # Exact slice mask of cone or cylindrical aperture through the cube. `weight_mask` is
        # a 3D array with fractions of each pixel within an aperture at each
        # wavelength, on the range [0, 1].
        if self.aperture.selected == self.aperture.default_text:
            # Entire Cube
            return self.inverted_mask_non_science

        return (
            self.inverted_mask_non_science *
            self.aperture.get_mask(
                self.dataset.selected_obj,
                self.aperture_method_selected,
                self.slice_display_unit,
                self.spatial_axes,
                self.reference_spectral_value if self.wavelength_dependent else None)
        )

    @property
    def bg_weight_mask(self):
        if self.background.selected == self.background.default_text:
            # NO background
            return np.zeros_like(self.dataset.selected_obj.flux.value)

        return (
            self.inverted_mask_non_science *
            self.background.get_mask(
                self.dataset.selected_obj,
                self.aperture_method_selected,
                self.slice_display_unit,
                self.spatial_axes,
                self.reference_spectral_value if self.bg_wavelength_dependent else None)
        )

    @property
    def aperture_area_along_spectral(self):
        # Weight mask summed along the spatial axes so that we get area of the aperture, in pixels,
        # as a function of wavelength.
        # To convert to steradians, multiply by self.cube.meta.get('PIXAR_SR', 1.0)
        return np.sum(self.aperture_weight_mask, axis=self.spatial_axes)

    @property
    def bg_area_along_spectral(self):
        return np.sum(self.bg_weight_mask, axis=self.spatial_axes)

    def _extract_from_aperture(self, cube, uncert_cube, mask_cube, aperture,
                               weight_mask, wavelength_dependent,
                               selected_func, **kwargs):
        # This plugin collapses over the *spatial axes* (optionally over a spatial subset,
        # defaults to ``No Subset``). Since the Cubeviz parser puts the fluxes
        # and uncertainties in different glue Data objects, we translate the spectral
        # cube and its uncertainties into separate NDDataArrays, then combine them:
        if not isinstance(aperture, ApertureSubsetSelect):
            raise ValueError("aperture must be an ApertureSubsetSelect object")
        if aperture.selected != aperture.default_text:
            nddata = cube.get_subset_object(
                subset_id=aperture.selected, cls=NDDataArray
            )
            if uncert_cube:
                uncertainties = uncert_cube.get_subset_object(
                    subset_id=aperture.selected, cls=StdDevUncertainty
                )
            else:
                uncertainties = None

            if self.aperture_method_selected.lower() == 'center':
                flux = nddata.data << nddata.unit
            else:  # exact (min/max not allowed here)
                # Apply the fractional pixel array to the flux cube
                flux = (weight_mask * nddata.data) << nddata.unit
            # Boolean cube which is True outside of the aperture
            # (i.e., the numpy boolean mask convention)
            mask = np.isclose(weight_mask, 0)

            # composite subset masks are in `nddata.mask`:
            if nddata.mask is not None and np.all(weight_mask == 0):
                mask &= nddata.mask

        else:
            nddata = cube.get_object(cls=NDDataArray)
            if uncert_cube:
                uncertainties = uncert_cube.get_object(cls=StdDevUncertainty)
            else:
                uncertainties = None
            flux = nddata.data << nddata.unit
            mask = nddata.mask

        # Use the spectral coordinate from the WCS:
        if '_orig_spec' in cube.meta:
            wcs = cube.meta['_orig_spec'].wcs.spectral
        elif hasattr(cube.coords, 'spectral'):
            wcs = cube.coords.spectral
        elif hasattr(cube.coords, 'spectral_wcs'):
            # This is the attribute for a PaddedSpectrumWCS in the 3D case
            wcs = cube.coords.spectral_wcs
        else:
            wcs = None

        # Filter out NaNs (False = good)
        mask = np.logical_or(mask, np.isnan(flux))

        # Also apply the cube's original mask array
        if mask_cube:
            snackbar_message = SnackbarMessage(
                "Note: Applied loaded mask cube during extraction",
                color="warning",
                sender=self)
            self.hub.broadcast(snackbar_message)
            mask_from_cube = mask_cube.get_component('flux').data.copy()
            # Some mask cubes have NaNs where they are not masked instead of 0
            mask_from_cube[np.where(np.isnan(mask_from_cube))] = 0
            mask = np.logical_or(mask, mask_from_cube.astype('bool'))

        nddata_reshaped = NDDataArray(
            flux, mask=mask, uncertainty=uncertainties, wcs=wcs, meta=nddata.meta
        )
        # by default we want to use operation_ignores_mask=True in nddata:
        kwargs.setdefault("operation_ignores_mask", True)
        # by default we want to propagate uncertainties:
        kwargs.setdefault("propagate_uncertainties", True)

        if selected_func == 'mean':
            # Use built-in sum function to collapse NDDataArray
            collapsed_sum_for_mean = nddata_reshaped.sum(axis=self.spatial_axes, **kwargs)
            # But we still need the mean function for everything except flux
            collapsed_as_mean = nddata_reshaped.mean(axis=self.spatial_axes, **kwargs)

            # Then normalize the flux based on the fractional pixel array
            flux_for_mean = (collapsed_sum_for_mean.data /
                             np.sum(weight_mask, axis=self.spatial_axes)) << nddata_reshaped.unit
            # Combine that information into a new NDDataArray
            collapsed_nddata = NDDataArray(flux_for_mean, mask=collapsed_as_mean.mask,
                                           uncertainty=collapsed_as_mean.uncertainty,
                                           wcs=collapsed_as_mean.wcs,
                                           meta=collapsed_as_mean.meta)
        elif selected_func == 'sum':
            collapsed_nddata = getattr(nddata_reshaped, selected_func)(
                axis=self.spatial_axes, **kwargs
            )  # returns an NDDataArray

            # Remove per solid angle denominator to turn sb into flux
            sq_angle_unit = check_if_unit_is_per_solid_angle(collapsed_nddata.unit,
                                                             return_unit=True)
            if sq_angle_unit is not None:
                # convert aperture area in steradians to the selected square angle unit
                # NOTE: just forcing these units for now!! this is in steradians and
                # needs to be converted to the selected square angle unit but for now just
                # force to correct units
                if sq_angle_unit == u.sr:
                    aperture_area = self.cube.meta.get('PIXAR_SR', 1.0) * sq_angle_unit
                else:
                    aperture_area = 1 * sq_angle_unit
                collapsed_nddata = collapsed_nddata.multiply(aperture_area,
                                                             propagate_uncertainties=True)
        else:
            collapsed_nddata = getattr(nddata_reshaped, selected_func)(
                axis=self.spatial_axes, **kwargs
            )  # returns an NDDataArray

        return self._return_extracted(cube, wcs, collapsed_nddata)

    def _return_extracted(self, cube, wcs, collapsed_nddata):
        # Convert to Spectrum1D, with the spectral axis in correct units:
        if hasattr(cube.coords, 'spectral_wcs'):
            target_wave_unit = cube.coords.spectral_wcs.world_axis_units[0]
        elif hasattr(cube.coords, 'spectral'):
            target_wave_unit = cube.coords.spectral.world_axis_units[0]
        else:
            target_wave_unit = None

        if target_wave_unit == '':
            target_wave_unit = 'pix'

        flux = collapsed_nddata.data << collapsed_nddata.unit
        mask = collapsed_nddata.mask
        uncertainty = collapsed_nddata.uncertainty

        collapsed_spec = _return_spectrum_with_correct_units(
            flux, wcs, collapsed_nddata.meta, data_type='flux',
            target_wave_unit=target_wave_unit,
            uncertainty=uncertainty,
            mask=mask
        )
        return collapsed_spec

    def _preview_x_from_extracted(self, extracted):
        return extracted.spectral_axis

    def _preview_y_from_extracted(self, extracted):
        """
        Convert y-axis units of extraction preview to display units,
        if necessary.
        """

        if extracted.flux.unit != self.spectrum_y_units:

            eqv = all_flux_unit_conversion_equivs(self.dataset.selected_obj.meta.get('PIXAR_SR', 1.0),  # noqa
                                                  self.dataset.selected_obj.spectral_axis)

            return flux_conversion_general(extracted.flux.value, extracted.flux.unit,
                                           self.spectrum_y_units, eqv)

        return extracted.flux

    @with_spinner()
    def extract(self, return_bg=False, add_data=True, **kwargs):
        """
        Extract the spectrum from the data cube according to the plugin inputs.

        Parameters
        ----------
        return_bg : bool, optional
            Whether to also return the spectrum of the background, if applicable.
        add_data : bool, optional
            Whether to load the resulting data back into the application according to
            ``add_results``.
        kwargs : dict
            Additional keyword arguments passed to the NDDataArray collapse operation.
            Examples include ``propagate_uncertainties`` and ``operation_ignores_mask``.
        """
        if self.conflicting_aperture_and_function:
            raise ValueError(self.conflicting_aperture_error_message)
        if self.aperture.selected == self.background.selected:
            raise ValueError("aperture and background cannot be set to the same subset")

        selected_func = self.function_selected.lower()
        spec = self._extract_from_aperture(self.cube, self.uncert_cube, self.mask_cube,
                                           self.aperture, self.aperture_weight_mask,
                                           self.wavelength_dependent,
                                           selected_func, **kwargs)

        bg_spec = self.extract_bg_spectrum(add_data=False, bg_spec_per_spaxel=False)
        if bg_spec is not None:
            spec = spec - bg_spec

        # per https://jwst-docs.stsci.edu/jwst-near-infrared-camera/nircam-performance/nircam-absolute-flux-calibration-and-zeropoints # noqa
        pix_scale_factor = self.cube.meta.get('PIXAR_SR', 1.0)
        spec.meta['_pixel_scale_factor'] = pix_scale_factor

        # stuff for exporting to file
        self.extracted_spec = spec
        self.extraction_available = True
        fname_label = self.dataset_selected.replace("[", "_").replace("]", "")
        self.filename = f"extracted_{selected_func}_{fname_label}.fits"

        if add_data:
            if default_color := self.aperture.selected_item.get('color', None):
                spec.meta['_default_color'] = default_color
            self.add_results.add_results_from_plugin(spec)

            snackbar_message = SnackbarMessage(
                f"{self.resulting_product_name.title()} extracted successfully.",
                color="success",
                sender=self)
            self.hub.broadcast(snackbar_message)

        if return_bg:
            return spec, bg_spec
        return spec

    @with_spinner('bg_spec_spinner')
    def extract_bg_spectrum(self, add_data=False, **kwargs):
        """
        Create a background 1D spectrum from the input parameters defined in the plugin.

        If ``function`` is 'sum', then the value is scaled by the relative ratios of the area
        (along the spectral axis) of ``aperture`` to ``background``.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting spectrum to the application, according to the options
            defined in the plugin.
        kwargs : dict
            Additional keyword arguments passed to the NDDataArray collapse operation.
            Examples include ``propagate_uncertainties`` and ``operation_ignores_mask``.
        """
        # allow internal calls to override the behavior of the bg_spec_per_spaxel traitlet
        bg_spec_per_spaxel = kwargs.pop('bg_spec_per_spaxel', self.bg_spec_per_spaxel)
        if self.background.selected != self.background.default_text:
            bg_spec = self._extract_from_aperture(self.cube, self.uncert_cube, self.mask_cube,
                                                  self.background, self.bg_weight_mask,
                                                  self.bg_wavelength_dependent,
                                                  self.function_selected.lower(), **kwargs)
            if self.function_selected.lower() == 'sum':
                if not bg_spec_per_spaxel:
                    # then scale according to aperture areas across the spectral axis (allowing for
                    # independent wavelength-dependence btwn the aperture and background)
                    bg_spec *= self.aperture_area_along_spectral / self.bg_area_along_spectral
        else:
            bg_spec = None

        if add_data:
            if bg_spec is None:
                raise ValueError(f"Background is set to {self.background.selected}")
            self.bg_spec_add_results.add_results_from_plugin(bg_spec, replace=False)

        return bg_spec

    def vue_spectral_extraction(self, *args, **kwargs):
        try:
            self.extract(add_data=True)
        except Exception as e:
            self.hub.broadcast(SnackbarMessage(
                f"Extraction failed: {repr(e)}",
                sender=self, color="error"))

    def vue_create_bg_spec(self, *args, **kwargs):
        self.extract_bg_spectrum(add_data=True)

    @observe('aperture_selected', 'function_selected')
    def _set_default_results_label(self, event={}):
        if not hasattr(self, 'aperture'):
            return
        if self.aperture.selected == self.aperture.default_text:
            self.results_label_default = f"{self.resulting_product_name.title()} ({self.function_selected.lower()})"  # noqa
        else:
            self.results_label_default = f"{self.resulting_product_name.title()} ({self.aperture_selected}, {self.function_selected.lower()})"  # noqa

    @cached_property
    def marks(self):
        if not self._tray_instance:
            return {}
        # TODO: iterate over self.slice_indicator_viewers and handle adding/removing viewers
        sv = self.slice_indicator_viewers[0]
        marks = {'extract': PluginLine(sv, visible=self.is_active),
                 'bg_extract': PluginLine(sv,
                                          line_style='dotted',
                                          visible=self.is_active and self.active_step == 'bg')}
        sv.figure.marks = sv.figure.marks + [marks['extract'], marks['bg_extract']]
        return marks

    def _clear_marks(self):
        for mark in self.marks.values():
            if mark.visible:
                mark.visible = False

    @observe('is_active', 'show_live_preview',
             'dataset_selected', 'aperture_selected', 'bg_selected',
             'wavelength_dependent', 'bg_wavelength_dependent', 'reference_spectral_value',
             'function_selected',
             'aperture_method_selected',
             'spectrum_y_units',
             'previews_temp_disabled')
    def _live_update_marks(self, event={}):
        if self.spectrum_y_units == '':
            # ensure that units are populated
            # which in turn will make a call back here
            # from the observe on spectrum_y_units
            self._on_global_display_unit_changed(None)
            return
        self._update_marks(event)

    @skip_if_not_tray_instance()
    def _update_marks(self, event={}):
        visible = self.show_live_preview and self.is_active

        if not visible:
            self._clear_marks()
            return

        # ensure the correct visibility, always (whether or not there have been updates)
        self.marks['bg_extract'].visible = self.active_step == 'bg' and self.bg_selected != self.background.default_text  # noqa
        self.marks['extract'].visible = True

        # _live_update_extract will skip if no updates since last active
        self._live_update_extract(event)

    @skip_if_no_updates_since_last_active()
    @with_temp_disable(timeout=0.4)
    def _live_update_extract(self, event={}):
        self._update_extract()

    @skip_if_not_tray_instance()
    def _update_extract(self):
        try:
            ext, bg_extract = self.extract(return_bg=True, add_data=False)
        except (ValueError, Exception):
            self._clear_marks()
            return False

        self.marks['extract'].update_xy(self._preview_x_from_extracted(ext),
                                        self._preview_y_from_extracted(ext))

        if bg_extract is None:
            self.marks['bg_extract'].clear()
            self.marks['bg_extract'].visible = False
        else:
            self.marks['bg_extract'].update_xy(self._preview_x_from_extracted(bg_extract),
                                               self._preview_y_from_extracted(bg_extract))



import pytest
from astropy import units as u
from astropy.coordinates import Angle, SkyCoord
from astropy.tests.helper import assert_quantity_allclose
from glue.core.roi import CircularROI, EllipticalROI, RectangularROI
from numpy.testing import assert_allclose
from photutils.aperture import (CircularAperture, SkyCircularAperture,
                                EllipticalAperture, SkyEllipticalAperture,
                                RectangularAperture, SkyRectangularAperture,
                                CircularAnnulus, SkyCircularAnnulus,
                                EllipticalAnnulus, SkyEllipticalAnnulus,
                                RectangularAnnulus, SkyRectangularAnnulus)
from regions import (CirclePixelRegion, EllipsePixelRegion, RectanglePixelRegion,
                     CircleAnnulusPixelRegion, EllipseAnnulusPixelRegion,
                     RectangleAnnulusPixelRegion, PolygonPixelRegion, PixCoord)

from jdaviz.core.region_translators import (
    regions2roi, regions2aperture, aperture2regions, PHOTUTILS_LT_2_2)


# TODO: Use proper method from upstream when that is available.
def compare_region_shapes(reg1, reg2):
    assert reg1.__class__ == reg2.__class__
    for param in reg1._params:
        par1 = getattr(reg1, param)
        par2 = getattr(reg2, param)
        if isinstance(par1, PixCoord):
            assert_allclose(par1.xy, par2.xy)
        elif isinstance(par1, SkyCoord):
            assert par1 == par2
        elif isinstance(par1, u.Quantity):
            assert_quantity_allclose(par1, par2)
        else:
            assert_allclose(par1, par2)


def test_translation_circle(image_2d_wcs):
    region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, CircularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r, region_shape.radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r, region_sky.radius)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    with pytest.raises(ValueError, match='regions shape only accepts scalar x and y positions'):
        aperture2regions(CircularAperture(((10, 20), (30, 40)), 3))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        CirclePixelRegion(center=PixCoord(x=[0, 42], y=[1, 43]), radius=4.2)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CirclePixelRegion(center=PixCoord(x=42, y=43), radius=[1, 4.2])


def test_translation_circle_roi(image_2d_wcs):
    region_shape = CirclePixelRegion(center=PixCoord(x=42, y=43), radius=4.2)
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, CircularROI)
    assert_allclose((roi_pix.xc, roi_pix.yc, roi_pix.radius), (42, 43, 4.2))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, CircularROI)
    assert_allclose((roi_sky.xc, roi_sky.yc, roi_sky.radius),
                    (roi_pix.xc, roi_pix.yc, roi_pix.radius))

    with pytest.raises(ValueError, match='WCS must be provided'):
        regions2roi(region_sky)


def test_translation_ellipse(image_2d_wcs):
    region_shape = EllipsePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, EllipticalAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a * 2, region_shape.width)
    assert_allclose(aperture.b * 2, region_shape.height)
    if PHOTUTILS_LT_2_2:
        assert_allclose(aperture.theta, region_shape.angle.radian)
    else:
        assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a * 2, region_sky.width)
    assert_quantity_allclose(aperture_sky.b * 2, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        EllipsePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), width=16, height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=[1, 16], height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=[1, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipsePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle([0, 30], 'deg'))


def test_translation_ellipse_roi(image_2d_wcs):
    region_shape = EllipsePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, EllipticalROI)
    assert_allclose((roi_pix.xc, roi_pix.yc, roi_pix.radius_x, roi_pix.radius_y, roi_pix.theta),
                    (42, 43, 8.0, 5.0, 0.5235987755982988))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, EllipticalROI)
    assert_allclose((roi_sky.xc, roi_sky.yc, roi_sky.radius_x, roi_sky.radius_y, roi_sky.theta),
                    (roi_pix.xc, roi_pix.yc, roi_pix.radius_x, roi_pix.radius_y, roi_pix.theta))


def test_translation_rectangle(image_2d_wcs):
    region_shape = RectanglePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, RectangularAperture)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w, region_shape.width)
    assert_allclose(aperture.h, region_shape.height)
    if PHOTUTILS_LT_2_2:
        assert_allclose(aperture.theta, region_shape.angle.radian)
    else:
        assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAperture)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w, region_sky.width)
    assert_quantity_allclose(aperture_sky.h, region_sky.height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        RectanglePixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), width=16, height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=[1, 16], height=10, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=[1, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectanglePixelRegion(
            center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle([0, 30], 'deg'))


def test_translation_rectangle_roi(image_2d_wcs):
    region_shape = RectanglePixelRegion(
        center=PixCoord(x=42, y=43), width=16, height=10, angle=Angle(30, 'deg'))
    roi_pix = regions2roi(region_shape)
    assert isinstance(roi_pix, RectangularROI)
    assert_allclose((roi_pix.xmin, roi_pix.xmax, roi_pix.ymin, roi_pix.ymax, roi_pix.theta),
                    (34, 50, 38, 48, 0.5235987755982988))

    region_sky = region_shape.to_sky(image_2d_wcs)
    roi_sky = regions2roi(region_sky, wcs=image_2d_wcs)
    assert isinstance(roi_sky, RectangularROI)
    assert_allclose((roi_sky.xmin, roi_sky.xmax, roi_sky.ymin, roi_sky.ymax, roi_sky.theta),
                    (roi_pix.xmin, roi_pix.xmax, roi_pix.ymin, roi_pix.ymax, roi_pix.theta))


def test_translation_circle_annulus(image_2d_wcs):
    region_shape = CircleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=8)
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, CircularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.r_in, region_shape.inner_radius)
    assert_allclose(aperture.r_out, region_shape.outer_radius)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyCircularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.r_in, region_sky.inner_radius)
    assert_quantity_allclose(aperture_sky.r_out, region_sky.outer_radius)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_radius=5, outer_radius=8)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=[1, 5], outer_radius=8)
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        CircleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_radius=5, outer_radius=[8, 10])


def test_translation_ellipse_annulus(image_2d_wcs):
    region_shape = EllipseAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
        outer_height=6.5, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, EllipticalAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.a_in * 2, region_shape.inner_width)
    assert_allclose(aperture.a_out * 2, region_shape.outer_width)
    assert_allclose(aperture.b_in * 2, region_shape.inner_height)
    assert_allclose(aperture.b_out * 2, region_shape.outer_height)
    if PHOTUTILS_LT_2_2:
        assert_allclose(aperture.theta, region_shape.angle.radian)
    else:
        assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyEllipticalAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.a_in * 2, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.a_out * 2, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.b_in * 2, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.b_out * 2, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_width=5.5, inner_height=3.5,
            outer_width=8.5, outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=[1, 5.5], inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=[1, 3.5], outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=[8.5, 10],
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=[6.5, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        EllipseAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle([0, 30], 'deg'))


def test_translation_rectangle_annulus(image_2d_wcs):
    region_shape = RectangleAnnulusPixelRegion(
        center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
        outer_height=6.5, angle=Angle(30, 'deg'))
    aperture = regions2aperture(region_shape)
    assert isinstance(aperture, RectangularAnnulus)
    assert_allclose(aperture.positions, region_shape.center.xy)
    assert_allclose(aperture.w_in, region_shape.inner_width)
    assert_allclose(aperture.w_out, region_shape.outer_width)
    assert_allclose(aperture.h_in, region_shape.inner_height)
    assert_allclose(aperture.h_out, region_shape.outer_height)
    if PHOTUTILS_LT_2_2:
        assert_allclose(aperture.theta, region_shape.angle.radian)
    else:
        assert_quantity_allclose(aperture.theta, region_shape.angle)

    region_sky = region_shape.to_sky(image_2d_wcs)
    aperture_sky = regions2aperture(region_sky)
    assert isinstance(aperture_sky, SkyRectangularAnnulus)
    assert aperture_sky.positions == region_sky.center  # SkyCoord
    assert_quantity_allclose(aperture_sky.w_in, region_sky.inner_width)
    assert_quantity_allclose(aperture_sky.w_out, region_sky.outer_width)
    assert_quantity_allclose(aperture_sky.h_in, region_sky.inner_height)
    assert_quantity_allclose(aperture_sky.h_out, region_sky.outer_height)
    assert_quantity_allclose(aperture_sky.theta + (90 * u.deg), region_sky.angle)

    # Roundtrip
    compare_region_shapes(region_shape, aperture2regions(aperture))
    compare_region_shapes(region_sky, aperture2regions(aperture_sky))

    # NOTE: If these no longer fail, we also have to account for non-scalar inputs.
    # Assume this is representative for the sky counterpart too.
    with pytest.raises(ValueError, match='must be a scalar PixCoord'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=[0, 42], y=[1, 43]), inner_width=5.5, inner_height=3.5,
            outer_width=8.5, outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=[1, 5.5], inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=[1, 3.5], outer_width=8.5,
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=[8.5, 10],
            outer_height=6.5, angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=[6.5, 10], angle=Angle(30, 'deg'))
    with pytest.raises(ValueError, match=r'must be .* scalar'):
        RectangleAnnulusPixelRegion(
            center=PixCoord(x=42, y=43), inner_width=5.5, inner_height=3.5, outer_width=8.5,
            outer_height=6.5, angle=Angle([0, 30], 'deg'))


def test_translation_polygon():
    region_shape = PolygonPixelRegion(vertices=PixCoord(x=[1, 2, 2], y=[1, 1, 2]))
    with pytest.raises(NotImplementedError, match='is not supported'):
        regions2aperture(region_shape)
    with pytest.raises(NotImplementedError, match='is not supported'):
        regions2roi(region_shape)



from importlib import resources
import json

from astropy.table import QTable

__all__ = ['get_linelist_metadata', 'get_available_linelists', 'load_preset_linelist']


def get_linelist_metadata():
    """Return metadata for line lists."""
    metadata_file = resources.files("jdaviz").joinpath("data/linelists/linelist_metadata.json")
    with open(metadata_file) as f:
        metadata = json.load(f)
    return metadata


def get_available_linelists():
    """
    Return all available line lists.

    Filters out all entries that do not explicitly contain medium information, as to not
    mislead users
    """
    metadata = get_linelist_metadata()
    return [list for list in list(metadata.keys()) if 'medium' in metadata[list]]


def load_preset_linelist(name):
    """Return one of the preset line lists, loaded into `~astropy.table.QTable`.
    """
    metadata = get_linelist_metadata()
    if name not in metadata.keys():
        raise ValueError("Line name not in available set of line lists. " +
                         "Valid list names are: {}".format(list(metadata.keys())))
    fname_base = metadata[name]["filename_base"]
    fname = resources.files("jdaviz").joinpath("data/linelists/{}.csv".format(fname_base))
    units = metadata[name]["units"]
    linetable = QTable.read(fname)

    # Add units
    linetable['Rest Value'].unit = units

    # Add column with list name reference
    linetable['listname'] = name

    # Rename remaining columns
    linetable.rename_columns(('Line Name', 'Rest Value'), ('linename', 'rest'))

    return linetable



from regions import Regions
from specutils import SpectralRegion


from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToPlugin


__all__ = ['SubsetImporter']


@loader_importer_registry('Subset')
class SubsetImporter(BaseImporterToPlugin):
    template_file = __file__, "subset.vue"

    @property
    def is_valid(self):
        return (isinstance(self.input, (Regions, SpectralRegion))
                and self.has_default_plugin)

    @property
    def default_plugin(self):
        return 'Subset Tools'

    def __call__(self, subset_label=None):
        self.app._jdaviz_helper.plugins['Subset Tools'].import_region(self.input)



import pytest
import numpy as np
from astropy import units as u
from astropy.table import QTable
from astropy.tests.helper import assert_quantity_allclose
from numpy.testing import assert_allclose
from regions import RectanglePixelRegion, PixCoord
from specutils import Spectrum1D, SpectralRegion
from glue.core.roi import XRangeROI

from jdaviz.configs.specviz.plugins.line_analysis.line_analysis import _coerce_unit
from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.core.events import LineIdentifyMessage
from jdaviz.core.marks import LineAnalysisContinuum


def test_plugin(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # disabling keep_active should hide the continuum
    plugin.keep_active = False
    assert np.all([cm.visible is False for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    assert 'Subset 1' in plugin.spectral_subset.labels
    plugin.selected_subset = 'Subset 1'
    plugin.selected_continuum = 'Surrounding'
    plugin.width = 3

    for result_dict in plugin.results:
        if result_dict in ['Line Flux']:
            # should have an assigned uncertainty (with min required version of specutils)
            assert len(result_dict.get('uncertainty')) > 0


@pytest.mark.filterwarnings(r"ignore:'W/m2/m' contains multiple slashes")
@pytest.mark.filterwarnings('ignore:No observer defined on WCS')
def test_spatial_subset(cubeviz_helper, image_cube_hdu_obj):
    """
    Tests that the spatial selection returns any valid result, DOES NOT VALIDATE THE VALUE
    Not checking the value attempts to circumvent the issue we ran into here:
    https://github.com/spacetelescope/jdaviz/pull/1564#discussion_r949427663
    """
    cubeviz_helper.load_data(image_cube_hdu_obj, data_label="Test Cube")

    # add a spatial region
    cubeviz_helper.plugins['Subset Tools'].import_region(
        RectanglePixelRegion(center=PixCoord(x=3, y=5), width=4, height=6))

    # create a spectral region
    unit = u.Unit(cubeviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    cubeviz_helper.plugins['Subset Tools'].combination_mode = 'new'
    cubeviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(3.623e-7 * unit,
                                                                        3.627e-7 * unit))
    cubeviz_helper.app.state.drawer_content = 'plugins'

    plugin = cubeviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    plugin.dataset_selected = 'Spectrum (Subset 1, sum)'
    plugin.spectral_subset_selected = 'Subset 2'
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.width = 1

    for result in plugin.results:
        # Check that there exists a value
        assert not np.isnan(float(result['result']))
        # Check the unit is not dimensionless
        assert u.Unit(result['unit']) != u.dimensionless_unscaled


@pytest.mark.parametrize('sq_angle_unit', [u.sr, PIX2])
def test_cubeviz_units(cubeviz_helper, spectrum1d_cube_custom_fluxunit,
                       sq_angle_unit):
    """
    Tests that the plugin produces the correct result when loaded cube
    is in flux/pix2 and flux/sr, and that the results remain consistant
    between translations of the spectral y axis flux<>surface brightness.
    """
    cube = spectrum1d_cube_custom_fluxunit(fluxunit=u.Jy / sq_angle_unit,
                                           shape=(25, 25, 4), with_uncerts=True)
    cubeviz_helper.load_data(cube, data_label="Test Cube")

    uc = cubeviz_helper.plugins['Unit Conversion']
    assert uc.spectral_y_type == 'Flux'  # initial selection should be Flux

    plugin = cubeviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    plugin.dataset_selected = 'Spectrum (sum)'
    plugin.continuum_subset_selected = 'Surrounding'

    results = plugin.results
    assert results[0]['unit'] == 'W / m2'

    viewer = cubeviz_helper.app.get_viewer('spectrum-viewer')
    viewer.apply_roi(XRangeROI(4.63e-7, 4.64e-7))

    la = cubeviz_helper.plugins['Line Analysis']
    la.keep_active = True
    la.spectral_subset.selected = 'Subset 1'

    marks_before = [la._obj.continuum_marks['left'].y,
                    la._obj.continuum_marks['right'].y]

    # change flux unit and make sure result stays the same after conversion
    uc.flux_unit.selected = 'MJy'

    marks_after = [la._obj.continuum_marks['left'].y,
                   la._obj.continuum_marks['right'].y]

    # ensure continuum marks update when spectral_y is changed by
    # multiply converted continuum marks by expected scale factor (MJy -> Jy)
    scaling_factor = 1e-6
    assert_allclose([mark * scaling_factor for mark in marks_before], marks_after, rtol=1e-5)

    # reset to test again after spectral_y_type is changed
    marks_before = marks_after

    # now change to surface brightness
    uc.spectral_y_type = 'Surface Brightness'

    if sq_angle_unit == PIX2:
        # translation does not alter spectral_y values in viewer
        scaling_factor = 1
    else:
        scaling_factor = cube.meta.get('PIXAR_SR')

    marks_after = [la._obj.continuum_marks['left'].y,
                   la._obj.continuum_marks['right'].y]

    # ensure continuum marks update when spectral_y_type is changed
    # multiply converted continuum marks by expected pixel scale factor
    assert_allclose([mark / scaling_factor for mark in marks_before], marks_after, rtol=1e-5)

    results = plugin.results
    line_flux_before_unit_conversion = results[0]
    # convert back and forth between unit<>str for string format consistency
    unit_str = u.Unit(f'W / {sq_angle_unit.to_string()} m2').to_string()
    assert line_flux_before_unit_conversion['unit'] == unit_str

    # change flux unit and make sure result stays the same after conversion
    uc.flux_unit.selected = 'MJy'
    results = plugin.results
    np.testing.assert_allclose(float(results[0]['result']),
                               float(line_flux_before_unit_conversion['result']))
    np.testing.assert_allclose(float(results[0]['uncertainty']),
                               float(line_flux_before_unit_conversion['uncertainty']))


def test_user_api(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))

    la = specviz_helper.plugins['Line Analysis']
    la.keep_active = True

    # spectral subset now supports multiselect
    assert "multiselect" in la.spectral_subset.__repr__()
    with pytest.raises(ValueError):
        la.spectral_subset.select_all()
    with pytest.raises(ValueError):
        la.spectral_subset.select_none()

    # test that setting a string to a Selection component redirects to the selected traitlet
    la.spectral_subset = 'Subset 1'
    assert la.spectral_subset.selected == 'Subset 1'

    # test that invalid strings are caught and reverted to the original selection
    with pytest.raises(ValueError):
        la.spectral_subset = 'invalid'

    assert la.spectral_subset.selected == 'Subset 1'

    la.get_results()


def test_line_identify(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    lt = QTable()
    lt['linename'] = ['O III', 'Halpha']
    lt['rest'] = [5007, 6563]*u.AA
    lt['listname'] = 'Test List'
    specviz_helper.load_line_list(lt)

    ll_plugin = specviz_helper.app.get_tray_item_from_name('g-line-list')
    la_plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    la_plugin.plugin_opened = True
    rest_names = [line['name_rest'] for line in ll_plugin.list_contents['Test List']['lines']]

    # will default to no selection
    assert la_plugin.selected_line == ''

    # check redshift
    assert la_plugin.selected_line_redshift == 0.0

    # but selecting a line from line-list (or clicking) should change the dropdown value
    # since sync is enabled by default
    assert la_plugin.sync_identify is True

    msg = LineIdentifyMessage(rest_names[1],
                              sender=specviz_helper)
    specviz_helper.app.session.hub.broadcast(msg)
    assert la_plugin.selected_line == rest_names[1]

    # and changing the dropdown should change the identified line
    la_plugin.selected_line = rest_names[0]
    assert ll_plugin.list_contents['Test List']['lines'][0].get('identify') is True
    assert ll_plugin.list_contents['Test List']['lines'][1].get('identify') is False

    # unless we disable the sync
    la_plugin.sync_identify = False
    la_plugin.selected_line = rest_names[1]
    assert ll_plugin.list_contents['Test List']['lines'][0].get('identify') is True
    assert ll_plugin.list_contents['Test List']['lines'][1].get('identify') is False

    # selected_line should become the same as identified_line
    assert la_plugin.selected_line == "Halpha 6563.0"
    assert la_plugin.identified_line == "O III 5007.0"
    la_plugin.sync_identify = True
    assert la_plugin.selected_line == "O III 5007.0"
    assert la_plugin.identified_line == "O III 5007.0"

    # identified_line should become the same as selected_line
    la_plugin.identified_line = ''
    la_plugin.sync_identify = False
    la_plugin.sync_identify = True
    assert la_plugin.identified_line == "O III 5007.0"

    # manually update redshift
    la_plugin.vue_line_assign()
    assert_allclose(la_plugin.results_centroid, 7307.4232674401555)
    line_mark = la_plugin.line_marks[la_plugin.line_items.index(la_plugin.selected_line)]
    assert_allclose(line_mark.rest_value, 5007)
    z = la_plugin._compute_redshift_for_selected_line()
    assert_allclose(z, (la_plugin.results_centroid - line_mark.rest_value)/line_mark.rest_value)
    assert_allclose(la_plugin.selected_line_redshift, z)


def test_coerce_unit():
    q_input = 1 * u.Unit('1E-20 erg m / (Angstrom cm**2 s)')
    q_input.uncertainty = 0.1 * u.Unit('1E-20 erg m / (Angstrom cm**2 s)')
    q_unit = u.Unit('erg / (cm**2 s)')
    q_coerced = _coerce_unit(q_input)
    assert_quantity_allclose(q_coerced, 1e-10 * q_unit)
    assert_quantity_allclose(q_coerced.uncertainty, 1e-11 * q_unit)
    q_input.uncertainty = None
    q_coerced = _coerce_unit(q_input)
    assert not hasattr(q_coerced, 'uncertainty')


def test_continuum_surrounding_spectral_subset(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.spectral_subset_selected = 'Subset 1'
    plugin.width = 3

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), 2.153181e-13, atol=1e-15)


def test_continuum_spectral_same_value(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels
    plugin.continuum_subset_selected = 'Subset 1'
    plugin.spectral_subset_selected = 'Subset 1'
    plugin.width = 3

    # continuum and spectral cannot be the same value
    assert plugin.results[0]['result'] == ''


def test_continuum_surrounding_invalid_width(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.spectral_subset_selected = 'Subset 1'
    plugin.continuum_width = 11
    assert plugin.results[0]['result'] == ''


def test_continuum_subset_spectral_entire(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7400 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels
    plugin.continuum_subset_selected = 'Subset 1'
    plugin.spectral_subset_selected = 'Entire Spectrum'
    plugin.width = 3

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), -2.79572e-13, atol=1e-15)


def test_continuum_subset_spectral_subset2(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6200 * unit,
                                                                        7000 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    specviz_helper.plugins['Subset Tools'].combination_mode = 'new'
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(7100 * unit,
                                                                        7700 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert plugin.spectral_subset.labels == ['Entire Spectrum', 'Subset 1', 'Subset 2']

    plugin.spectral_subset_selected = 'Subset 2'
    plugin.continuum_subset_selected = 'Subset 1'
    plugin.width = 3

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), 1.482418e-14, atol=1e-16)


def test_continuum_surrounding_no_right(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        8000 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels

    plugin.spectral_subset_selected = 'Subset 1'
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.width = 3

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), 4.204513e-14, atol=1e-16)


def test_continuum_surrounding_no_left(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6000 * unit,
                                                                        7500 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels

    plugin.spectral_subset_selected = 'Subset 1'
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.width = 3

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), 7.570859e-14, atol=1e-16)


def test_subset_changed(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    plugin.keep_active = True

    # continuum should be created, plotted, and visible
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    continuum_marks = [m for m in sv.figure.marks if isinstance(m, LineAnalysisContinuum)]
    assert len(continuum_marks) == 3
    assert np.all([cm.visible for cm in continuum_marks])

    # add a region and rerun stats for that region
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6000 * unit,
                                                                        7500 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    plugin = specviz_helper.app.get_tray_item_from_name('specviz-line-analysis')
    assert 'Subset 1' in plugin.spectral_subset.labels

    plugin.spectral_subset_selected = 'Subset 1'
    plugin.continuum_subset_selected = 'Surrounding'
    plugin.width = 3

    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(6500 * unit,
                                                                        7500 * unit))
    specviz_helper.app.state.drawer_content = 'plugins'

    # Values have not yet been validated
    np.testing.assert_allclose(float(plugin.results[0]['result']), 2.153181e-13, atol=1e-15)


def test_invalid_subset(specviz_helper, spectrum1d):
    # 6000-8000
    specviz_helper.load_data(spectrum1d, data_label="right_spectrum")

    # 5000-7000
    sp2 = Spectrum1D(spectral_axis=spectrum1d.spectral_axis - 1000*spectrum1d.spectral_axis.unit,
                     flux=spectrum1d.flux * 1.25)
    specviz_helper.load_data(sp2, data_label="left_spectrum")

    # apply subset that overlaps on left_spectrum, but not right_spectrum
    # NOTE: using a subset that overlaps the right_spectrum (reference) results in errors when
    # retrieving the subset (https://github.com/spacetelescope/jdaviz/issues/1868)
    unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    specviz_helper.plugins['Subset Tools'].import_region(SpectralRegion(5000 * unit,
                                                                        6000 * unit))

    plugin = specviz_helper.plugins['Line Analysis']
    plugin.dataset = 'right_spectrum'
    assert plugin.dataset == 'right_spectrum'
    assert plugin.spectral_subset == 'Entire Spectrum'
    assert plugin._obj.spectral_subset_valid

    plugin.spectral_subset = 'Subset 1'
    assert not plugin._obj.spectral_subset_valid

    with pytest.raises(ValueError, match=r"spectral subset 'Subset 1' \(5000.0, 6000.0\) is outside data range of 'right_spectrum' \(6000.0, 8000.0\)"):  # noqa
        plugin.get_results()

    plugin.dataset = 'left_spectrum'
    assert plugin._obj.spectral_subset_valid
